import asyncio
import os
import wave

from unittest import TestCase

import numpy as np

from wafl.config import Configuration
from wafl.connectors.bridges.llm_chitchat_answer_bridge import LLMChitChatAnswerBridge
from wafl.connectors.local.local_llm_connector import LocalLLMConnector
from wafl.connectors.remote.remote_llm_connector import RemoteLLMConnector
from wafl.listener.whisper_listener import WhisperListener
from wafl.speaker.fairseq_speaker import FairSeqSpeaker

_path = os.path.dirname(__file__)


class TestConnection(TestCase):
    def test__connection_to_generative_model_can_generate_text(self):
        config = Configuration.load_local_config()
        connector = RemoteLLMConnector(config.get_value("llm_model"))
        prediction = asyncio.run(
            connector.predict(
                'Generate a full paragraph based on this chapter title "The first contact". '
                "The theme of the paragraph is space opera. "
            )
        )
        assert len(prediction) > 0

    def test__connection_to_generative_model_can_generate_text_within_tags(self):
        config = Configuration.load_local_config()
        connector = RemoteLLMConnector(config.get_value("llm_model"))
        connector._num_prediction_tokens = 200
        text = 'Generate a full paragraph based on this chapter title " The First Contact". The theme of the paragraph is space opera. Include the characters "Alberto" and "Maria". Write at least three sentences.'
        prompt = f"""
<task>
Complete the following task and add <|EOS|> at the end: {text}
</task>
<result>
                """.strip()

        prediction = asyncio.run(connector.predict(prompt))
        print(prediction)
        assert len(prediction) > 0

    def test__connection_to_generative_model_can_generate_a_python_list(self):
        config = Configuration.load_local_config()
        connector = RemoteLLMConnector(config.get_value("llm_model"))
        connector._num_prediction_tokens = 200
        prompt = "Generate a Python list of 4 chapters names for a space opera book. The output needs to be a python list of strings: "
        prediction = asyncio.run(connector.predict(prompt))
        print(prediction)
        assert len(prediction) > 0

    def test__local_llm_connector_can_generate_a_python_list(self):
        config = Configuration.load_from_filename("local_config.json")
        connector = LocalLLMConnector(config.get_value("llm_model"))
        connector._num_prediction_tokens = 200
        prompt = "Generate a list of 4 chapters names for a space opera book. The output needs to be a python list of strings: "
        prediction = asyncio.run(connector.predict(prompt))
        print(prediction)
        assert len(prediction) > 0

    def test__chit_chat_bridge_can_run_locally(self):
        config = Configuration.load_from_filename("local_config.json")
        dialogue_bridge = LLMChitChatAnswerBridge(config)
        answer = asyncio.run(dialogue_bridge.get_answer("", "", "bot: hello"))
        assert len(answer) > 0

    def test__listener_local_connector(self):
        config = Configuration.load_from_filename("local_config.json")
        listener = WhisperListener(config)
        f = wave.open(os.path.join(_path, "data/1002.wav"), "rb")
        waveform = np.frombuffer(f.readframes(f.getnframes()), dtype=np.int16) / 32768
        result = asyncio.run(listener.input_waveform(waveform))
        expected = "DELETE BATTERIES FROM THE GROCERY LIST"
        assert expected.lower() in result

    def test__speaker_local_connector(self):
        config = Configuration.load_from_filename("local_config.json")
        speaker = FairSeqSpeaker(config)
        text = "Hello world"
        asyncio.run(speaker.speak(text))



import re


def _change_code_wrapper(text):
    pattern = r"```.*?\n(.*?)```"
    num_rows = text.count("\n") - 1

    def replace_code(match):
        code = match.group(1)
        return f'<textarea readonly id="code" rows="{num_rows}">{code}</textarea>'

    return re.sub(pattern, replace_code, text, flags=re.DOTALL)


def get_html_from_dialogue_item(
    text,
):
    if text.find("bot:") == 0:
        return (
            f"<div class='shadow-lg dialogue-row-bot' style='font-size:20px; '"
            f">" + _change_code_wrapper(text[4:]).strip() + "</div>"
        )

    return (
        "<div class='dialogue-row-user' style='font-size:20px; '>"
        + text[5:].strip()
        + "</div>"
    )



import base64
import aiohttp
import asyncio
import json

from typing import Dict


class RemoteSpeakerConnector:
    _max_tries = 3

    def __init__(self, config):
        self._server_url = (
            f"https://{config['model_host']}:"
            f"{config['model_port']}/predictions/speaker"
        )
        try:
            loop = asyncio.get_running_loop()

        except RuntimeError:
            loop = None

        if (not loop or (loop and not loop.is_running())) and not asyncio.run(
            self.check_connection()
        ):
            raise RuntimeError("Cannot connect a running Speaker Model.")

    async def predict(self, text: str) -> Dict[str, float]:
        payload = {"text": text}
        for _ in range(self._max_tries):
            async with aiohttp.ClientSession(
                connector=aiohttp.TCPConnector(ssl=False)
            ) as session:
                async with session.post(self._server_url, json=payload) as response:
                    data = await response.text()
                    prediction = json.loads(data)
                    wav = prediction["wav"]
                    rate = prediction["rate"]
                    return {"wav": base64.b64decode(wav), "rate": rate}

        return {"wav": "", "rate": 16000}

    async def check_connection(self):
        payload = {"text": "test"}
        try:
            async with aiohttp.ClientSession(
                conn_timeout=3, connector=aiohttp.TCPConnector(ssl=False)
            ) as session:
                async with session.post(self._server_url, json=payload) as response:
                    await response.text()
                    return True

        except aiohttp.client.InvalidURL:
            print()
            print("Is the whisper server running?")
            print("Please run 'bash start-llm.sh' (see docs for explanation).")
            print()

        return False



import decimal

from num2words import num2words


def convert_numbers_to_words(text):
    words = text.split()
    new_words = []
    for word in words:
        new_word = word
        try:
            if any(number in word for number in "1234567890"):
                new_word = num2words(word)

        except decimal.InvalidOperation as e:
            pass

        new_words.append(new_word)

    return " ".join(new_words).replace("-", " ")



from typing import List, Any
from wafl.events.BaseEventCreator import BaseEventsCreator


class EventsCreatorFromFunctionList(BaseEventsCreator):
    def __init__(self, functions: List[Any]):
        self._functions = functions

    def get(self) -> List[str]:
        return [function() for function in self._functions]



from wafl.simple_text_processing.questions import is_question


class Narrator:
    def __init__(self, interface):
        self._interface = interface

    def summarize_dialogue(self):
        dialogue_list = self._interface.get_utterances_list()[-7:]
        summary = ""
        for line in dialogue_list:
            speaker = self.get_speaker(line)
            utterance = self.get_utterance(line)
            if not speaker or not utterance:
                continue

            if is_question(utterance):
                summary += f"when the {speaker} asks: '{utterance}' "

            else:
                summary += f"the {speaker} says: '{utterance}'; "

        summary = self._clean_summary(summary)
        return summary

    def get_speaker(self, line):
        if ":" not in line:
            return None

        return line.split(":")[0]

    def get_utterance(self, line):
        if ":" not in line:
            return None

        return line.split(":")[1]

    def get_context_for_facts(self, text):
        text = text.strip()
        return f"The bot remembers: '{text}'"

    def get_relevant_query_answer_context(self, text, query_text, answer):
        if "the user says:" in text.lower():
            return f"when asked: '{query_text}' the user says '{answer}'"

        if "the bot says:" in text.lower():
            return f"when asked: '{query_text}' the bot says '{answer}'"

        if "the bot remembers:" in text.lower():
            return f"when asked: '{query_text}' the bot remembers '{answer}'"

        return f"the answer to '{query_text}' is '{answer}'"

    def get_relevant_fact_context(self, text, answer):
        text = text.strip()
        answer = answer.strip()
        if "the user says:" in text.lower():
            return f"the user says '{answer}'"

        if "the bot says:" in text.lower():
            return f"the bot says '{answer}'"

        if "the bot remembers:" in text.lower():
            return f"the bot remembers '{answer}'"

        return answer

    def _clean_summary(self, text):
        text = text.replace(";.", ";")
        return text



import asyncio
import math

import pyaudio
import time
import numpy as np

from wafl.connectors.factories.whisper_connector_factory import WhisperConnectorFactory


class WhisperListener:
    _chunk = 1024
    _format = pyaudio.paInt16
    _channels = 1
    _rate = 16000
    _range = 32768
    _generation_max_length = 15
    _starting_tokens = [50257, 50362]
    _ending_tokens = [50256]

    def __init__(self, config):
        self._p = pyaudio.PyAudio()
        self._volume_threshold = 1
        self._original_volume_threshold = self._volume_threshold
        self._timeout = 1
        self._max_timeout = 4
        self._hotword_threshold = -8
        self._connector = WhisperConnectorFactory.get_connector(config)
        self._hotwords = list()
        self.is_active = False
        self._last_waveform = None

    def set_hotwords(self, hotwords):
        self._hotwords = [item.lower() for item in hotwords]

    def add_hotwords(self, hotwords):
        if hotwords and not type(hotwords) == list:
            hotwords = [hotwords]

        self._hotwords.extend([item.lower() for item in hotwords])

    def set_timeout(self, timeout):
        self._timeout = timeout

    def set_volume_threshold(self, threshold):
        self._volume_threshold = threshold
        self._original_volume_threshold = self._volume_threshold

    def set_hotword_threshold(self, threshold):
        self._hotword_threshold = threshold

    def record(self, start_with):
        rec = list()
        rec.append(start_with)

        current = time.time()
        end = time.time() + self._timeout
        upper_limit_end = time.time() + self._max_timeout

        while current <= end and current < upper_limit_end:
            data = self.stream.read(self._chunk)
            if _rms(data) >= self._volume_threshold:
                end = time.time() + self._timeout

            current = time.time()
            rec.append(data)

        return np.frombuffer(b"".join(rec), dtype=np.int16) / self._range

    def activate(self):
        if not self.is_active:
            self.stream = self._p.open(
                format=self._format,
                channels=self._channels,
                rate=self._rate,
                input=True,
                output=False,
                frames_per_buffer=self._chunk,
            )
            self.is_active = True

    def deactivate(self):
        if self.is_active:
            self.stream.stop_stream()
            self.stream.close()
            self.is_active = False

    async def input(self):
        if not self.is_active:
            self.activate()

        while True:
            await asyncio.sleep(0)
            inp = self.stream.read(self._chunk)
            rms_val = _rms(inp)
            if rms_val > self._volume_threshold:
                waveform = self.record(start_with=inp)
                self.deactivate()
                return await self.input_waveform(waveform)

            else:
                new_threshold = 2 * rms_val
                self._volume_threshold = max(
                    new_threshold, self._original_volume_threshold
                )

    async def input_waveform(self, waveform):
        self._last_waveform = waveform
        prediction = await self._connector.predict(waveform)
        transcription = prediction["transcription"]
        score = prediction["score"]

        if math.exp(score) > 0.5:
            return transcription

        return "[unclear]"

    async def get_hotword_if_present(self):
        for hotword in self._hotwords:
            if await self.hotword_is_present(hotword):
                return hotword

        return ""

    async def hotword_is_present(self, hotword):
        if type(self._last_waveform) != np.ndarray:
            raise RuntimeError(
                "The waveform has not been processed. Please call input_waveform() before hotword_is_present()"
            )
        prediction = await self._connector.predict(self._last_waveform, hotword=hotword)
        return prediction["logp"] > self._hotword_threshold


def _rms(frame):
    data = np.frombuffer(frame, dtype=np.int16)
    return np.std(data) / len(data)



import csv
import os
import joblib

from wafl.knowledge.single_file_knowledge import SingleFileKnowledge

_path = os.path.dirname(__file__)


async def load_knowledge_from_file(filename, config):
    items_list = []
    with open(os.path.join(_path, "../../data/", filename + ".csv")) as file:
        csvreader = csv.reader(file)
        for row in csvreader:
            items_list.append(row[0].strip())

    knowledge = await SingleFileKnowledge.create_from_list(items_list, config)
    joblib.dump(knowledge, os.path.join(_path, f"../../data/{filename}.knowledge"))
    return knowledge



import os
import sys

from wafl.config import create_initial_files
from wafl.parsing.preprocess import remove_preprocessed
from wafl.run import (
    run_from_command_line,
    run_testcases,
    print_incipit,
    download_models,
)
from wafl.runners.run_from_audio import run_from_audio
from wafl.runners.run_web_interface import run_app


def print_help():
    print("\n")
    print("These are the available commands:")
    print("> wafl init: Initialize the current folder")
    print("> wafl run-cli: Run a cli version of the chatbot")
    print("> wafl run-audio: Run a voice-powered version of the chatbot")
    print("> wafl run-server: Run a webserver version of the chatbot")
    print("> wafl run-tests: Run the tests in testcases.txt")
    print()


def add_cwd_to_syspath():
    sys.path.append(os.getcwd())


def process_cli():
    add_cwd_to_syspath()
    print_incipit()

    arguments = sys.argv
    if len(arguments) > 1:
        command = arguments[1]

        if command == "init":
            create_initial_files()
            download_models()

        elif command == "run-cli":
            run_from_command_line()
            remove_preprocessed("/")

        elif command == "run-audio":
            run_from_audio()
            remove_preprocessed("/")

        elif command == "run-server":
            run_app()
            remove_preprocessed("/")

        elif command == "run-tests":
            run_testcases()
            remove_preprocessed("/")

        elif command == "help":
            print_help()

        else:
            print("Unknown argument.\n")
    else:
        print_help()


def main():
    try:
        process_cli()

    except RuntimeError as e:
        print(e)
        print("WAFL ended due to the exception above.")



def normalized(text, lower_case=True):
    text = text.strip()
    if not text:
        return ""

    if text[-1] == ".":
        text = text[:-1]

    if lower_case:
        text = text.lower()

    return text.strip()



from wafl.config import Configuration
from wafl.events.conversation_events import ConversationEvents
from wafl.interface.voice_interface import VoiceInterface
from wafl.logger.local_file_logger import LocalFileLogger
from wafl.scheduler.conversation_loop import ConversationLoop
from wafl.scheduler.scheduler import Scheduler

_logger = LocalFileLogger()


def run_from_audio():
    config = Configuration.load_local_config()
    interface = VoiceInterface(config)
    conversation_events = ConversationEvents(
        config=config,
        interface=interface,
        logger=_logger,
    )
    conversation_loop = ConversationLoop(
        interface,
        conversation_events,
        _logger,
        activation_word=config.get_value("waking_up_word"),
    )
    scheduler = Scheduler([conversation_loop])
    scheduler.run()


if __name__ == "__main__":
    run_from_audio()



import asyncio


class GeneratedEventLoop:
    def __init__(self, interface, events, logger):
        self._interface = interface
        self._events = events
        self._logger = logger
        self._repeat_every_seconds = 60

    async def run(self):
        try:
            while True:
                await self._events.process_next()
                await asyncio.sleep(self._repeat_every_seconds)

        except (KeyboardInterrupt, EOFError) as e:
            self._logger.write(
                f"The Generated event loop has ended due to exception {str(e)}"
            )



from wafl.connectors.remote.remote_speaker_connector import RemoteSpeakerConnector


class SpeakerConnectorFactory:
    @staticmethod
    def get_connector(config):
        return RemoteSpeakerConnector(config.get_value("speaker_model"))



from wafl.connectors.remote.remote_llm_connector import RemoteLLMConnector


class LLMConnectorFactory:
    @staticmethod
    def get_connector(config):
        return RemoteLLMConnector(config.get_value("llm_model"))



import time

from typing import List


class BaseInterface:
    def __init__(self, decorator=None):
        self._is_listening = True
        self._choices = []
        self._facts = []
        self._utterances = []
        self._decorator = decorator

    async def output(self, text: str, silent: bool = False):
        raise NotImplementedError

    async def input(self) -> str:
        raise NotImplementedError

    def bot_has_spoken(self, to_set: bool = None):
        raise NotImplementedError

    def is_listening(self):
        return self._is_listening

    def activate(self):
        self._is_listening = True

    def deactivate(self):
        self._is_listening = False
        self._choices = []
        self._facts = []
        self._utterances = []

    def add_hotwords(self, hotwords: List[str]):
        raise NotImplementedError

    async def add_choice(self, text):
        self._choices.append((time.time(), text))
        await self.output(f"Making the choice: {text}", silent=True)

    async def add_fact(self, text):
        self._facts.append((time.time(), text))
        await self.output(f"{text}", silent=True)

    def get_choices_and_timestamp(self):
        return self._choices

    def get_facts_and_timestamp(self):
        return self._facts

    def get_utterances_list(self):
        return [item[1] for item in self._utterances]

    def get_utterances_list_with_timestamp(self):
        return self._utterances

    def reset_history(self):
        self._utterances = []
        self._choices = []
        self._facts = []

    def _decorate_reply(self, text: str) -> str:
        if not self._decorator:
            return text

        return self._decorator.extract(text, self._utterances)



from unittest import TestCase
from wafl.config import Configuration
from wafl.interface.voice_interface import VoiceInterface

_wafl_greetings = """
  
""".strip()


class TestConfig(TestCase):
    def test__listener_accepts_threshold_for_hotword_logp(self):
        config = Configuration.load_local_config()
        interface = VoiceInterface(config)
        self.assertEqual(
            interface._listener._hotword_threshold,
            config.get_value("listener_model")["listener_hotword_logp"],
        )

    def test__listener_accepts_threshold_for_volume(self):
        config = Configuration.load_local_config()
        interface = VoiceInterface(config)
        self.assertEqual(
            interface._listener._volume_threshold,
            config.get_value("listener_model")["listener_volume_threshold"],
        )

    def test__listener_accepts_silence_timeout(self):
        config = Configuration.load_local_config()
        interface = VoiceInterface(config)
        self.assertEqual(
            interface._listener._timeout,
            config.get_value("listener_model")["listener_silence_timeout"],
        )



import asyncio
import os

from unittest import TestCase

from wafl.config import Configuration
from wafl.speaker.fairseq_speaker import FairSeqSpeaker
from wafl.speaker.soundfile_speaker import SoundFileSpeaker

_wafl_greetings = """
  
""".strip()

_path = os.path.dirname(__file__)


class TestSpeaker(TestCase):
    def test_voice(self):
        config = Configuration.load_local_config()
        speaker = FairSeqSpeaker(config)
        text = "Hello world"
        asyncio.run(speaker.speak(text))

    def test_long_text(self):
        config = Configuration.load_local_config()
        speaker = FairSeqSpeaker(config)
        text = (
            "Shall I compare you to a summer's day? Thou art more lovely and temperate."
        )
        asyncio.run(speaker.speak(text))

    def test_number_pronunciation(self):
        config = Configuration.load_local_config()
        speaker = FairSeqSpeaker(config)
        text = "The time is 54 past 8"
        asyncio.run(speaker.speak(text))

    def test_on_sound(self):
        speaker = SoundFileSpeaker()
        speaker.speak(os.path.join(_path, "../wafl/sounds/activation.wav"))

    def test_off_sound(self):
        speaker = SoundFileSpeaker()
        speaker.speak(os.path.join(_path, "../wafl/sounds/deactivation.wav"))



import aiohttp
import asyncio

from wafl.connectors.base_llm_connector import BaseLLMConnector


class RemoteLLMConnector(BaseLLMConnector):
    _max_tries = 3
    _max_reply_length = 1024
    _num_prediction_tokens = 200
    _cache = {}
    _num_replicas = 10

    def __init__(self, config, last_strings=None):
        super().__init__(last_strings)
        host = config["model_host"]
        port = config["model_port"]
        self._server_url = f"https://{host}:{port}/predictions/bot"

        try:
            loop = asyncio.get_running_loop()

        except RuntimeError:
            loop = None

        if (not loop or (loop and not loop.is_running())) and not asyncio.run(
            self.check_connection()
        ):
            raise RuntimeError("Cannot connect a running LLM.")

    async def predict(self, prompt: str, temperature=None, num_tokens=None) -> [str]:
        if not temperature:
            temperature = 0.5

        if not num_tokens:
            num_tokens = self._num_prediction_tokens

        payload = {
            "data": prompt,
            "temperature": temperature,
            "num_tokens": num_tokens,
            "last_strings": self._last_strings,
            "num_replicas": self._num_replicas,
        }

        for _ in range(self._max_tries):
            async with aiohttp.ClientSession(
                connector=aiohttp.TCPConnector(ssl=False)
            ) as session:
                async with session.post(self._server_url, json=payload) as response:
                    answer = await response.text()
                    return answer.split("<||>")

        return [""]

    async def check_connection(self):
        payload = {
            "data": "test",
            "temperature": 0.6,
            "num_tokens": 1,
            "last_strings": self._last_strings,
            "num_replicas": self._num_replicas,
        }
        try:
            async with aiohttp.ClientSession(
                conn_timeout=3, connector=aiohttp.TCPConnector(ssl=False)
            ) as session:
                async with session.post(self._server_url, json=payload) as response:
                    await response.text()
                    return True

        except aiohttp.client.InvalidURL:
            print()
            print("Is the wafl-llm running?")
            print("Please run 'bash start-llm.sh' (see docs for explanation).")
            print()

        return False



from importlib import import_module
from inspect import getmembers, isfunction
from typing import List
from wafl.events.BaseEventCreator import BaseEventsCreator


class EventsCreatorFromModuleName(BaseEventsCreator):
    def __init__(self, module_name):
        self._module = {}
        self._functions = {}
        self._module = import_module(module_name)
        self._functions = [item[0] for item in getmembers(self._module, isfunction)]

    def get(self) -> List[str]:
        return [
            eval(f"self._module.{function}()", {"self": self})
            for function in self._functions
        ]



from wafl.parsing.utils import get_lines_stripped_from_comments

_user_prompt = "user:"
_bot_prompt = "bot:"


def get_user_and_bot_lines_from_text(text: str):
    lines = get_lines_stripped_from_comments(text)

    testcase_name = ""
    testcases = {}
    to_negate = False
    bot_lines = []
    user_lines = []
    all_lines = []

    for line in lines:
        if not line.strip():
            continue

        separation = line.find(line.strip())
        if separation == 0:
            if user_lines or bot_lines:
                testcases[testcase_name] = {
                    "bot_lines": bot_lines,
                    "user_lines": user_lines,
                    "lines": all_lines,
                    "negated": to_negate,
                }
                to_negate = False
                bot_lines = []
                user_lines = []
                all_lines = []

            line = line.strip()
            if line.find("!") == 0:
                to_negate = True
                line = line[1:].strip()

            testcase_name = line
            testcases[testcase_name] = {}
            continue

        line = line.strip()
        all_lines.append(line)

        if line.find(_user_prompt) == 0:
            user_lines.append(line[len(_user_prompt) :].strip())
            continue

        if line.find(_bot_prompt) == 0:
            bot_lines.append(line[len(_bot_prompt) :].strip())
            continue

    if user_lines or bot_lines:
        testcases[testcase_name] = {
            "bot_lines": bot_lines,
            "user_lines": user_lines,
            "lines": all_lines,
            "negated": to_negate,
        }

    return testcases



import asyncio
import random
import traceback

from wafl.exceptions import CloseConversation


class ConversationLoop:
    def __init__(
        self,
        interface,
        conversation,
        logger,
        activation_word="",
        max_misses=3,
        deactivate_on_closed_conversation=True,
    ):
        self._interface = interface
        self._conversation_events = conversation
        self._logger = logger
        self._activation_word = activation_word
        self._max_misses = max_misses
        self._deactivate_on_closed_conversation = deactivate_on_closed_conversation

    async def run(self):
        await self._say_initial_greetings()
        try:
            await self._main_loop()

        except (KeyboardInterrupt, EOFError):
            await self._interface.output("Good bye!")

    async def _say_initial_greetings(self):
        self._interface.activate()
        if self._activation_word:
            await self._interface.output(
                f"Please say '{self._activation_word}' to activate me"
            )
            self._interface.add_hotwords(self._activation_word)
            self._interface.deactivate()

    async def _main_loop(self):
        max_misses = self._max_misses
        num_misses = 0
        interactions = 0
        while True:
            await asyncio.sleep(0)
            if not self._interface.is_listening():
                interactions = 0
                num_misses = 0

            try:
                result = await self._conversation_events.process_next(
                    activation_word=self._activation_word
                )
                self._logger.write(
                    f"Conversation Result {result}", log_level=self._logger.level.INFO
                )
                interactions += 1
                if result:
                    self._interface.activate()

                if (
                    self._activation_word
                    and self._interface.bot_has_spoken()
                    and interactions == 1
                ):
                    self._interface.deactivate()
                    num_misses = 0

                if (
                    self._interface.is_listening()
                    and not result
                    and not self._interface.bot_has_spoken()
                ):
                    num_misses += 1
                    if max_misses > 0 and num_misses >= max_misses:
                        self._interface.deactivate()
                        self._interface.reset_history()

                    if interactions == 1:
                        await self._interface.output(
                            random.choice(["What can I do for you?"])
                        )

                    else:
                        await self._interface.output(
                            random.choice(["Sorry?", "Can you repeat?"])
                        )

                else:
                    num_misses = 0

            except CloseConversation:
                self._logger.write(
                    f"Closing the conversation", log_level=self._logger.level.INFO
                )
                if self._deactivate_on_closed_conversation:
                    self._interface.deactivate()

                else:
                    await self._interface.output("I am not sure what to reply.")

                continue

            except Exception as e:
                self._logger.write(
                    "Error in conversation loop", log_level=self._logger.level.ERROR
                )
                self._logger.write(str(e), log_level=self._logger.level.ERROR)
                print("Error in conversation loop. Closing the conversation.")
                await self._interface.output(
                    "Error in the conversation loop. Marking this conversation as a failure."
                )
                print(str(e))
                traceback.print_stack()



import re
import time

from wafl.simple_text_processing.deixis import from_bot_to_user
from wafl.interface.base_interface import BaseInterface
from wafl.interface.utils import not_good_enough


class DummyInterface(BaseInterface):
    def __init__(self, to_utter=None, output_filter=None):
        super().__init__()
        self._to_utter = to_utter
        self._bot_has_spoken = False
        self._dialogue = ""
        self._output_filter = output_filter

    async def output(self, text: str, silent: bool = False):
        if silent:
            print(text)
            return

        if self._output_filter:
            text = await self._output_filter.filter(
                self.get_utterances_list_with_timestamp(), text
            )

        self._dialogue += "bot: " + text + "\n"
        self._utterances.append((time.time(), f"bot: {from_bot_to_user(text)}"))
        self.bot_has_spoken(True)

    async def input(self) -> str:
        text = self._to_utter.pop(0).strip()
        text = self.__remove_activation_word_and_normalize(text)
        while self._is_listening and not_good_enough(text):
            await self.output("I did not quite understand that")
            text = self._to_utter.pop(0)

        self._dialogue += "user: " + text + "\n"
        utterance = text
        self._utterances.append((time.time(), f"user: {utterance}"))
        return utterance

    def bot_has_spoken(self, to_set: bool = None):
        if to_set != None:
            self._bot_has_spoken = to_set

        return self._bot_has_spoken

    def get_dialogue(self):
        return self._dialogue

    def __remove_activation_word_and_normalize(self, text):
        activation_word = re.sub(r"\[(.*)\].*", r"\1", text)
        text = re.sub(
            f"^\[{activation_word}\] {activation_word} (.*)",
            r"\1",
            text,
            flags=re.IGNORECASE,
        )
        return text



def select_best_answer(answers, last_strings):
    special_words = (
        last_strings
        + ["</remember>", "</execute>", "result ="]
        + ["<execute>", "<remember>", "<execute>", "<remember>"]
    )
    return sorted(
        answers, key=lambda x: sum([x.count(word) for word in special_words])
    )[-1]



import asyncio

from unittest import TestCase

from wafl.config import Configuration
from wafl.events.conversation_events import ConversationEvents
from wafl.exceptions import CloseConversation
from wafl.interface.dummy_interface import DummyInterface

wafl_example = """
rules:
  - the user thanks the bot:
    - The intention of the user is to close the conversation
    - You must answer the user by writing "<execute>close_conversation()</execute>"
"""


class TestInterruptionsToCloseConversation(TestCase):
    def test__thank_you_closes_conversation(self):
        interface = DummyInterface(
            to_utter=[
                "thank you",
            ]
        )
        config = Configuration.load_local_config()
        config.set_value("rules", wafl_example)
        conversation_events = ConversationEvents(
            config=config,
            interface=interface,
        )
        try:
            asyncio.run(conversation_events.process_next())

        except CloseConversation:
            self.assertTrue(True)
            return

        self.assertTrue(False)



class BaseKnowledge:
    root_knowledge = "/"

    async def add(self, text):
        raise NotImplementedError()

    async def add_rule(self, rule_text, knowledge_name=None):
        raise NotImplementedError()

    async def ask_for_facts(self, query, is_from_user=False, knowledge_name=None):
        raise NotImplementedError()

    async def ask_for_facts_with_threshold(
        self, query, is_from_user=False, knowledge_name=None, threshold=None
    ):
        raise NotImplementedError()

    async def ask_for_rule_backward(self, query, knowledge_name=None, first_n=None):
        raise NotImplementedError()

    async def has_better_match(self, query_text: str) -> bool:
        raise NotImplementedError()



import re

from wafl.knowledge.single_file_knowledge import SingleFileKnowledge
from wafl.simple_text_processing.normalize import normalized


def input_is_valid(text):
    if not text.strip():
        return False

    if normalized(text) != "no" and len(text) <= 2:
        return False

    return True


def remove_text_between_brackets(text: str) -> str:
    return re.sub(r"(\[.*?\])", "", text)


def load_knowledge(config, logger):
    if ".yaml" in config.get_value("rules") and not any(
        item in config.get_value("rules") for item in [" ", "\n"]
    ):
        with open(config.get_value("rules")) as file:
            rules_txt = file.read()

    else:
        rules_txt = config.get_value("rules")

    return SingleFileKnowledge(config, rules_txt, logger=logger)



import os
import random
import re
import time

from wafl.events.utils import remove_text_between_brackets
from wafl.simple_text_processing.deixis import from_bot_to_user
from wafl.interface.base_interface import BaseInterface
from wafl.interface.utils import get_most_common_words, not_good_enough
from wafl.listener.whisper_listener import WhisperListener
from wafl.speaker.fairseq_speaker import FairSeqSpeaker
from wafl.speaker.soundfile_speaker import SoundFileSpeaker

_path = os.path.dirname(__file__)


COLOR_START = "\033[94m"
COLOR_END = "\033[0m"


class VoiceInterface(BaseInterface):
    def __init__(self, config, output_filter=None):
        super().__init__()
        self._sound_speaker = SoundFileSpeaker()
        self._activation_sound_filename = self.__get_activation_sound_from_config(
            config
        )
        self._deactivation_sound_filename = self.__get_deactivation_sound_from_config(
            config
        )
        self.listener_model_name = config.get_value("listener_model")["local_model"]
        self._speaker = FairSeqSpeaker(config)
        self._listener = WhisperListener(config)
        self._listener.set_timeout(
            config.get_value("listener_model")["listener_silence_timeout"]
        )
        self._listener.set_volume_threshold(
            config.get_value("listener_model")["listener_volume_threshold"]
        )
        self._listener.set_hotword_threshold(
            config.get_value("listener_model")["listener_hotword_logp"]
        )
        self._bot_has_spoken = False
        self._utterances = []
        self._output_filter = output_filter

    async def add_hotwords_from_knowledge(
        self, knowledge: "Knowledge", max_num_words: int = 100, count_threshold: int = 5
    ):
        hotwords = get_most_common_words(
            knowledge.get_facts_and_rule_as_text(),
            max_num_words=max_num_words,
            count_threshold=count_threshold,
        )
        hotwords = [word.lower() for word in hotwords]
        self._listener.add_hotwords(hotwords)

    def add_hotwords(self, hotwords):
        self._listener.add_hotwords(hotwords)

    async def output(self, text: str, silent: bool = False):
        if silent:
            print(text)
            return

        if not text:
            return

        if self._output_filter:
            text = await self._output_filter.filter(
                self.get_utterances_list_with_timestamp(), text
            )

        self._listener.activate()
        text = from_bot_to_user(text)
        self._utterances.append((time.time(), f"bot: {text}"))
        print(COLOR_START + "bot> " + text + COLOR_END)
        await self._speaker.speak(text)
        self.bot_has_spoken(True)

    async def input(self) -> str:
        text = ""
        while not text:
            text = await self._listener.input()
            text = self.__remove_activation_word_and_normalize(text)
            hotword = await self._listener.get_hotword_if_present()
            if hotword:
                text = f"[{hotword}] {text}"

        while self._is_listening and not_good_enough(text):
            print(COLOR_START + "user> " + text + COLOR_END)
            await self.output(random.choice(["Sorry?", "Can you repeat?"]))
            text = await self._listener.input()

        text = text.lower().capitalize()
        print(COLOR_START + "user> " + text + COLOR_END)
        utterance = remove_text_between_brackets(text)
        if utterance.strip():
            self._utterances.append((time.time(), f"user: {text}"))

        return text

    def bot_has_spoken(self, to_set: bool = None):
        if to_set != None:
            self._bot_has_spoken = to_set

        return self._bot_has_spoken

    def activate(self):
        if not self._is_listening:
            self._sound_speaker.speak(self._activation_sound_filename)
            super().activate()

    def deactivate(self):
        if self._is_listening:
            self._sound_speaker.speak(self._deactivation_sound_filename)
            super().deactivate()

    def __get_activation_sound_from_config(self, config):
        if config.get_value("waking_up_sound"):
            return os.path.join(_path, "../sounds/activation.wav")

        return None

    def __get_deactivation_sound_from_config(self, config):
        if config.get_value("deactivate_sound"):
            return os.path.join(_path, "../sounds/deactivation.wav")

        return None

    def __remove_activation_word_and_normalize(self, text):
        activation_word = re.sub(r"\[(.*)\].*", r"\1", text)
        text = re.sub(
            f"^\[{activation_word}\] {activation_word} (.*)",
            r"\1",
            text,
            flags=re.IGNORECASE,
        )
        return text



import re
import time
import traceback

from importlib import import_module
from inspect import getmembers, isfunction

from wafl.answerer.answerer_implementation import (
    get_last_bot_utterances,
    get_last_user_utterance,
)
from wafl.answerer.base_answerer import BaseAnswerer
from wafl.connectors.bridges.llm_chitchat_answer_bridge import LLMChitChatAnswerBridge
from wafl.exceptions import CloseConversation
from wafl.extractors.dataclasses import Query, Answer
from wafl.simple_text_processing.questions import is_question


class DialogueAnswerer(BaseAnswerer):
    def __init__(self, config, knowledge, interface, code_path, logger):
        self._bridge = LLMChitChatAnswerBridge(config)
        self._knowledge = knowledge
        self._logger = logger
        self._interface = interface
        self._max_num_past_utterances = 5
        self._max_num_past_utterances_for_facts = 5
        self._max_num_past_utterances_for_rules = 0
        self._prior_facts_with_timestamp = []
        self._init_python_module(code_path.replace(".py", ""))
        self._max_predictions = 3

    async def answer(self, query_text):
        if self._logger:
            self._logger.write(f"Dialogue Answerer: the query is {query_text}")

        query = Query.create_from_text(query_text)
        rules_texts = await self._get_relevant_rules(query)

        dialogue = self._interface.get_utterances_list_with_timestamp()[
            -self._max_num_past_utterances :
        ]
        start_time = -1
        if dialogue:
            start_time = dialogue[0][0]

        if not dialogue:
            dialogue = [(time.time(), f"user: {query_text}")]

        dialogue_items = dialogue
        dialogue_items = sorted(dialogue_items, key=lambda x: x[0])
        last_bot_utterances = get_last_bot_utterances(dialogue_items, num_utterances=3)
        last_user_utterance = get_last_user_utterance(dialogue_items)
        dialogue_items = [item[1] for item in dialogue_items if item[0] >= start_time]
        conversational_timestamp = len(dialogue_items)
        facts = await self._get_relevant_facts(
            query,
            has_prior_rules=bool(rules_texts),
            conversational_timestamp=conversational_timestamp,
        )
        dialogue_items = "\n".join(dialogue_items)

        for _ in range(self._max_predictions):
            original_answer_text = await self._bridge.get_answer(
                text=facts,
                dialogue=dialogue_items,
                query=rules_texts,
            )
            await self._interface.add_fact(f"The bot predicts: {original_answer_text}")
            (
                answer_text,
                memories,
            ) = await self._substitute_memory_in_answer_and_get_memories_if_present(
                await self._substitute_results_in_answer(original_answer_text)
            )
            if answer_text in last_bot_utterances:
                dialogue_items = last_user_utterance
                continue

            if not memories:
                break

            facts += "\n" + "\n".join(memories)
            dialogue_items += f"\nbot: {original_answer_text}"

        if self._logger:
            self._logger.write(f"Answer within dialogue: The answer is {answer_text}")

        return Answer.create_from_text(answer_text)

    async def _get_relevant_facts(
        self, query, has_prior_rules, conversational_timestamp
    ):
        memory = "\n".join([item[0] for item in self._prior_facts_with_timestamp])
        self._prior_facts_with_timestamp = [
            item
            for item in self._prior_facts_with_timestamp
            if item[1]
            > conversational_timestamp - self._max_num_past_utterances_for_facts
        ]
        facts_and_thresholds = await self._knowledge.ask_for_facts_with_threshold(
            query, is_from_user=True, knowledge_name="/", threshold=0.8
        )
        if facts_and_thresholds:
            facts = [item[0].text for item in facts_and_thresholds if item[0].text not in memory]
            self._prior_facts_with_timestamp.extend(
                (item, conversational_timestamp) for item in facts
            )
            memory = "\n".join([item[0] for item in self._prior_facts_with_timestamp])

        else:
            if is_question(query.text) and not has_prior_rules:
                memory += (
                    f"\nThe answer to {query.text} is not in the knowledge base."
                    "The bot can answer the question while informing the user that the answer was not retrieved"
                )

        if has_prior_rules:
            memory += f"\nThe bot tries to answer {query.text} following the rules from the user."

        return memory

    async def _get_relevant_rules(self, query, max_num_rules=1):
        rules = await self._knowledge.ask_for_rule_backward(
            query,
            knowledge_name="/",
        )
        rules = rules[:max_num_rules]
        rules_texts = []
        for rule in rules:
            rules_text = f"- If {rule.effect.text} go through the following points:\n"
            for cause_index, causes in enumerate(rule.causes):
                rules_text += f"    {cause_index + 1}) {causes.text}\n"

            rules_texts.append(rules_text)
            await self._interface.add_fact(f"The bot remembers the rule:\n{rules_text}")

        return "\n".join(rules_texts)

    def _init_python_module(self, module_name):
        self._module = import_module(module_name)
        self._functions = [item[0] for item in getmembers(self._module, isfunction)]

    async def _substitute_results_in_answer(self, answer_text):
        matches = re.finditer(r"<execute>(.*?)</execute>", answer_text, re.DOTALL)
        for match in matches:
            to_execute = match.group(1)
            result = await self._run_code(to_execute)
            answer_text = answer_text.replace(match.group(0), result)

        return answer_text

    async def _substitute_memory_in_answer_and_get_memories_if_present(
        self, answer_text
    ):
        matches = re.finditer(r"<remember>(.*?)</remember>", answer_text, re.DOTALL)
        memories = []
        for match in matches:
            to_execute = match.group(1)
            answer_text = answer_text.replace(match.group(0), "")
            memories.append(to_execute)

        return answer_text, memories

    async def _run_code(self, to_execute):
        result = None
        for _ in range(3):
            try:
                if any(item + "(" in to_execute for item in self._functions):
                    result = eval(f"self._module.{to_execute}")
                    break

                else:
                    ldict = {}
                    exec(to_execute, globals(), ldict)
                    if "result" in ldict:
                        result = str(ldict["result"])
                        break

            except NameError as e:
                match = re.search(r"\'(\w+)\' is not defined", str(e))
                if match:
                    to_import = match.group(1)
                    to_execute = f"import {to_import}\n{to_execute}"

            except CloseConversation as e:
                raise e

            except Exception as e:
                result = (
                    f'Error while executing\n\n"""python\n{to_execute}\n"""\n\n{str(e)}'
                )
                traceback.print_exc()
                break

        if not result:
            result = f'\n"""python\n{to_execute}\n"""'

        return result



import asyncio
import os

from flask import render_template, request, jsonify
from wafl.interface.queue_interface import QueueInterface
from wafl.logger.history_logger import HistoryLogger
from wafl.scheduler.web_interface_implementation import (
    get_html_from_dialogue_item,
)

_path = os.path.dirname(__file__)


class WebLoop:
    def __init__(
        self,
        interface: QueueInterface,
        conversation_id: int,
        conversation_events: "ConversationEvents",
    ):
        self._interface = interface
        self._history_logger = HistoryLogger(self._interface)
        self._conversation_id = conversation_id
        self._conversation_events = conversation_events
        self._prior_dialogue_items = ""

    async def index(self):
        return render_template("index.html", conversation_id=self._conversation_id)

    async def handle_input(self):
        query = request.form["query"]
        self._interface.input_queue.append(query)
        return f"""
    <textarea id="query" type="text"
           class='shadow-lg'
           placeholder="{query}"
           name="query"
           hx-post="/{self._conversation_id}/input"
           hx-swap="outerHTML"
           hx-target="#query"
           hx-trigger="keydown[!shiftKey&&keyCode==13]"
    ></textarea>
        """.strip()

    async def reset_conversation(self):
        self._interface.reset_history()
        self._interface.deactivate()
        self._interface.activate()
        self._conversation_events.reload_knowledge()
        await self._interface.output("Hello. How may I help you?")
        conversation = await self._get_conversation()
        return conversation

    async def reload_rules(self):
        async with asyncio.Lock():
            print("Not implemented yet")

        return ""

    async def check_for_new_messages(self):
        conversation = await self._get_conversation()
        if conversation != self._prior_dialogue_items:
            self._prior_dialogue_items = conversation
            return f"""
            <div id="load_conversation" 
               hx-post="/{self._conversation_id}/load_messages"
               hx-swap="innerHTML"
               hx-target="#messages"
               hx-trigger="load"
            ></div>"""

        else:
            self._prior_dialogue_items = conversation
            return "<div id='load_conversation'></div>"

    async def load_messages(self):
        conversation = await self._get_conversation()
        return conversation

    async def handle_output(self):
        if not self._interface.output_queue:
            return jsonify({"text": "", "silent": False})

        output = self._interface.output_queue.pop(0)
        return jsonify(output)

    async def thumbs_up(self):
        self._history_logger.write("thumbs_up")
        return jsonify("")

    async def thumbs_down(self):
        self._history_logger.write("thumbs_down")
        return jsonify("")

    async def run(self):
        print(f"New web server instance {self._conversation_id} running!")
        return

    async def _get_conversation(self):
        dialogue_items = self._interface.get_utterances_list_with_timestamp()
        dialogue = []
        for index, item in enumerate(dialogue_items):
            dialogue.append(
                (
                    item[0],
                    get_html_from_dialogue_item(
                        item[1],
                    ),
                )
            )

        choices = self._interface.get_choices_and_timestamp()
        choices = [
            (
                item[0],
                "<div class='log-row'>" + item[1] + "</div>",
            )
            for item in choices
        ]
        facts = self._interface.get_facts_and_timestamp()
        facts = [
            (
                item[0],
                "<div class='log-row'>" + item[1] + "</div>",
            )
            for item in facts
        ]
        choices_and_facts = choices + facts
        choices_and_facts = sorted(choices_and_facts, key=lambda x: x[0])[::-1]
        choices_and_facts = [item[1] for item in choices_and_facts]
        dialogue_items = dialogue
        dialogue_items = sorted(dialogue_items, key=lambda x: x[0])[::-1]
        dialogue_items = [item[1] for item in dialogue_items]
        conversation = (
            "<div id='dialogue' class='dialogue overflow-y-scroll rounded-lg'>"
        )
        conversation += "".join(dialogue_items)
        conversation += "</div>"
        conversation += "<div id='logs' class='logs shadow-lg overflow-y-scroll rounded-lg width' style='flex-direction: column-reverse;'>"
        conversation += "".join(choices_and_facts)
        conversation += "</div>"
        return conversation



import functools
import importlib
import os
import re
from inspect import getmembers, isfunction

_preprocessed_prefix = "__wafl_"
_functions_standard_name = "functions"
_python_functions_standard_name = _functions_standard_name + ".py"


def get_all_functions_names(module_name):
    module = importlib.import_module(module_name)
    functions = [item[0] for item in getmembers(module, isfunction)]
    return functions


def clean_module_name(module_name):
    if not module_name:
        raise RuntimeError(
            f"The name {module_name} is empty and cannot be used as a python module"
        )

    names = module_name.split("/")
    return_path = []
    for name in names:
        if name and name != "." and name != "..":
            return_path.append(name)

        if name == "..":
            return_path.pop()

    if not return_path:
        return ""

    return_path = ".".join(return_path) + "."
    return return_path


@functools.lru_cache
def create_preprocessed(
    module: str,
    functions_standard_name=_functions_standard_name,
    python_functions_standard_name=_python_functions_standard_name,
):
    function_names = get_all_functions_names(
        clean_module_name(module) + functions_standard_name
    )
    filename = "." + module + "/" + python_functions_standard_name
    with open(filename) as file:
        print(f"Preprocessing {filename}.")
        text = file.read()
        text = re.sub(
            r'f?"%(.*)%"',
            'await inference.get_inference_answer(f"\\1", policy, task_memory)',
            text,
        )
        for name in function_names:
            text = re.sub(
                f" ({name}\([0-9a-zA-Z,\s:_]+)\)",
                " \\1, inference, policy, task_memory)",
                text,
            )
            text = re.sub(
                f" ({name})\(\)", " \\1(inference, policy, task_memory)", text
            )
            text = re.sub(
                f"(def {name}\()",
                "async \\1",
                text,
            )
            text = re.sub(f" ({name})\(", " await \\1(", text)
            text = re.sub(f'"[\s]*await ({name})\(', '"\\1(', text)

        text = text.replace("def await", "def")

    preprocessed_filename = filename.replace(
        _python_functions_standard_name,
        _preprocessed_prefix + _python_functions_standard_name,
    )
    with open(preprocessed_filename, "w") as file:
        file.write(text)


def import_module(
    module_name,
    preprocessed_prefix=_preprocessed_prefix,
    functions_standard_name=_functions_standard_name,
):
    module_name = (
        clean_module_name(module_name) + preprocessed_prefix + functions_standard_name
    )
    return importlib.import_module(module_name)


def remove_preprocessed(
    module, python_functions_standard_name=_python_functions_standard_name
):
    print("Removing preprocessed files")
    filename = (
        "." + module + "/" + _preprocessed_prefix + python_functions_standard_name
    )
    if os.path.isfile(filename):
        os.remove(filename)



import logging
import re


from wafl.connectors.utils import select_best_answer

_system_logger = logging.getLogger(__file__)

model = None
tokenizer = None


class BaseLLMConnector:
    _max_tries = 3
    _max_reply_length = 500
    _num_prediction_tokens = 200
    _cache = {}

    def __init__(self, last_strings=None):
        if not last_strings:
            self._last_strings = [
                "\nuser",
                "\nbot",
                "<|EOS|>",
                "</remember>",
                "</execute>\n",
                "</s>",
            ]

        else:
            self._last_strings = last_strings

    async def predict(self, prompt: str) -> [str]:
        raise NotImplementedError

    async def generate(self, prompt: str) -> str:
        if prompt in self._cache:
            return self._cache[prompt]

        text = prompt
        start = len(text)
        while (
            all(item not in text[start:] for item in self._last_strings)
            and len(text) < start + self._max_reply_length
        ):
            text += select_best_answer(await self.predict(text), self._last_strings)

        end_set = set()
        for item in self._last_strings:
            if "</remember>" in item or "</execute>" in item:
                continue

            end_set.add(text.find(item, start))

        if -1 in end_set:
            end_set.remove(-1)

        end = len(text)
        if end_set:
            end = min(end_set)

        candidate_answer = text[start:end].split("bot: ")[-1].strip()
        candidate_answer = re.sub(r"(.*)<\|.*\|>", r"\1", candidate_answer).strip()

        if prompt not in self._cache:
            self._cache[prompt] = candidate_answer

        if not candidate_answer:
            candidate_answer = "unknown"

        return candidate_answer



import asyncio

from unittest import TestCase

from wafl.config import Configuration
from wafl.events.conversation_events import ConversationEvents
from wafl.interface.dummy_interface import DummyInterface

wafl_example = """
facts:
  - the bots name is "Bob"
"""


class TestFacts(TestCase):
    def test__facts_are_retrieved(self):
        interface = DummyInterface(
            to_utter=[
                "what is your name",
            ]
        )
        config = Configuration.load_local_config()
        config.set_value("rules", wafl_example)
        conversation_events = ConversationEvents(
            config=config,
            interface=interface,
        )
        asyncio.run(conversation_events.process_next())
        print(interface.get_utterances_list())
        expected = "bob"
        self.assertIn(expected, interface.get_utterances_list()[-1].lower())



from dataclasses import dataclass


@dataclass
class Fact:
    text: str
    is_question: bool = False
    variable: str = None
    is_interruption: bool = False
    source: str = None
    destination: str = None
    knowledge_name: str = "/"

    def toJSON(self):
        return str(self)



import asyncio
import os
import wave
import numpy as np

from unittest import TestCase
from wafl.config import Configuration
from wafl.interface.voice_interface import VoiceInterface
from wafl.events.conversation_events import ConversationEvents
from wafl.interface.dummy_interface import DummyInterface
from wafl.listener.whisper_listener import WhisperListener

_wafl_example = """
facts:
  - This bot is doing well
  
rules:
  - the user's name is Jane:
    - write "I hear you"
""".strip()

_path = os.path.dirname(__file__)


class TestVoice(TestCase):
    def test__activation(self):
        interface = DummyInterface(to_utter=["computer", "my name is Jane"])
        config = Configuration.load_local_config()
        config.set_value("rules", _wafl_example)
        conversation_events = ConversationEvents(config=config, interface=interface)
        interface.activate()
        asyncio.run(conversation_events.process_next(activation_word="computer"))
        asyncio.run(conversation_events.process_next(activation_word="computer"))
        assert interface.get_utterances_list()[-1] == "bot: I hear you"

    def test__no_activation(self):
        interface = DummyInterface(to_utter=["my name is bob"])
        config = Configuration.load_local_config()
        conversation_events = ConversationEvents(config=config, interface=interface)
        interface.deactivate()
        asyncio.run(conversation_events.process_next(activation_word="computer"))
        assert len(interface.get_utterances_list()) == 1

    def test__computer_name_is_removed_after_activation(self):
        interface = DummyInterface(to_utter=["[computer] computer my name is bob"])
        config = Configuration.load_local_config()
        conversation_events = ConversationEvents(config=config, interface=interface)
        interface.deactivate()
        asyncio.run(conversation_events.process_next(activation_word="computer"))
        assert interface.get_utterances_list()[-1].count("computer") == 0

    def test__sound_file_is_translated_correctly(self):
        f = wave.open(os.path.join(_path, "data/1002.wav"), "rb")
        waveform = np.frombuffer(f.readframes(f.getnframes()), dtype=np.int16) / 32768
        config = Configuration.load_local_config()
        listener = WhisperListener(config)
        result = asyncio.run(listener.input_waveform(waveform))
        result = _normalize_utterance(result)
        expected = "DELETE BATTERIES FROM THE GROCERY LIST"
        assert result == expected

    def test__random_sounds_are_excluded(self):
        f = wave.open(os.path.join(_path, "data/random_sounds.wav"), "rb")
        waveform = np.frombuffer(f.readframes(f.getnframes()), dtype=np.int16) / 32768
        config = Configuration.load_local_config()
        listener = WhisperListener(config)
        result = asyncio.run(listener.input_waveform(waveform))
        expected = "[unclear]"
        assert result == expected

    def test__voice_interface_receives_config(self):
        config = Configuration.load_local_config()
        interface = VoiceInterface(config)
        assert (
            interface.listener_model_name
            == config.get_value("listener_model")["local_model"]
        )

    def test__hotword_listener_activated_using_recording_of_hotword(self):
        f = wave.open(os.path.join(_path, "data/computer.wav"), "rb")
        waveform = np.frombuffer(f.readframes(f.getnframes()), dtype=np.int16) / 32768
        config = Configuration.load_local_config()
        listener = WhisperListener(config)
        asyncio.run(listener.input_waveform(waveform))
        result = asyncio.run(listener.hotword_is_present("computer"))
        assert result

    def test__hotword_listener_is_not_activated_using_recording_of_not_hotword(self):
        f = wave.open(os.path.join(_path, "data/1002.wav"), "rb")
        waveform = np.frombuffer(f.readframes(f.getnframes()), dtype=np.int16) / 32768
        config = Configuration.load_local_config()
        listener = WhisperListener(config)
        asyncio.run(listener.input_waveform(waveform))
        result = asyncio.run(listener.hotword_is_present("computer"))
        assert not result


def _normalize_utterance(text):
    text = text.upper()
    for item in [".", ",", "!", ":", ";", "?"]:
        text = text.replace(item, " ")

    text = text.replace("  ", " ")
    return text.strip()



import os
import re

from wafl.events.answerer_creator import create_answerer
from wafl.simple_text_processing.normalize import normalized
from wafl.config import Configuration
from wafl.events.utils import input_is_valid, load_knowledge
from wafl.simple_text_processing.questions import is_question
from wafl.exceptions import InterruptTask

os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"


class ConversationEvents:
    def __init__(
        self,
        config: "Configuration",
        interface: "BaseInterface",
        logger=None,
    ):
        self._config = config
        self._knowledge = load_knowledge(config, logger)
        self._answerer = create_answerer(config, self._knowledge, interface, logger)
        self._interface = interface
        self._logger = logger
        self._is_computing = False
        if logger:
            self._logger.set_depth(0)

    async def output(self, text: str):
        await self._interface.output(text)

    async def _process_query(self, text: str):
        self._is_computing = True
        self._interface.bot_has_spoken(False)
        if not input_is_valid(text):
            self._is_computing = False
            return False

        text_is_question = is_question(text)
        try:
            answer = await self._answerer.answer(text)

        except InterruptTask:
            await self._interface.output("Task interrupted")
            self._is_computing = False
            return False

        if not self._interface.bot_has_spoken():
            if not text_is_question and not answer.is_neutral():
                await self.output(answer.text)

            if text_is_question and answer.text not in ["True", "False"]:
                await self.output(answer.text)

            if text_is_question and answer.is_false():
                await self.output("I don't know")

            if (
                not text_is_question
                and self._interface.get_utterances_list()[-1].find("user:") == 0
            ):
                await self._interface.output("I don't know what to reply")

            if (
                not text_is_question
                and answer.is_true()
                and not self._interface.bot_has_spoken()
            ):
                await self._interface.output("Yes")

        self._is_computing = False
        return answer

    async def process_next(self, activation_word: str = "") -> bool:
        try:
            text = await self._interface.input()
            text = text.replace("'", r"\'")

        except IndexError:
            return False

        if (
            activation_word
            and not self._interface.is_listening()
            and self._activation_word_in_text(activation_word, text)
        ):
            self._interface.activate()
            self._logger.set_depth(0)
            self._logger.write(f"Activation word found {text}")
            if normalized(text) == normalized(activation_word, lower_case=True):
                return True

        text = self.__remove_activation_word_and_normalize(activation_word, text)
        if self._interface.is_listening():
            answer = await self._process_query(text)
            if answer and not answer.is_false():
                return True

        return False

    def is_computing(self):
        return self._is_computing

    def reload_knowledge(self):
        self._knowledge = load_knowledge(self._config, self._logger)

    def _activation_word_in_text(self, activation_word, text):
        if f"[{normalized(activation_word)}]" in normalized(text):
            return True

        return False

    def __remove_activation_word_and_normalize(self, activation_word, text):
        to_remove = f"[{activation_word}]"
        text = text.replace(to_remove, "").strip()
        text = re.sub(f"^{activation_word}", "", text, flags=re.IGNORECASE)
        return text



import os

from unittest import TestCase
from wafl.logger.local_file_logger import LocalFileLogger

_path = os.path.dirname(__file__)


class TestLogger(TestCase):
    def test__setting_depth_to_zero(self):
        logger = LocalFileLogger()
        logger.set_depth(0)

    def test__setting_logger_level(self):
        logger = LocalFileLogger()
        logger.set_log_level(logger.level.INFO)

    def test__write_to_log(self):
        logger = LocalFileLogger()
        logger.write("Hello logs")



from typing import List, Tuple


class BaseRetriever:
    async def add_text_and_index(self, text: str, index: str):
        raise NotImplemented

    async def get_indices_and_scores_from_text(
        self, text: str
    ) -> List[Tuple[str, float]]:
        raise NotImplemented



def get_lines_stripped_from_comments(text):
    if not text:
        return []

    lines = []
    for line in text.split("\n"):
        if line.strip() and line.strip()[0] == "#":
            continue

        line = line.split("#")[0]
        lines.append(line)

    lines.append("")
    return lines


def is_quoted_text(text):
    if not text:
        return False

    if text[0] == '"' and text[-1] == '"':
        return True

    return False


def text_has_interruption(text):
    interruption_text = "INTERRUPTION"
    if text.find(interruption_text) == 0:
        return True

    return False


def clean_text(text):
    return text.replace("INTERRUPTION ", "")


def concatenate_slashes_into_one_single_line(text):
    text = text.replace("\\n", "")
    return text



from fuzzywuzzy import fuzz
from wafl.events.utils import load_knowledge
from wafl.simple_text_processing.deixis import from_user_to_bot, from_bot_to_user
from wafl.exceptions import CloseConversation
from wafl.events.conversation_events import ConversationEvents
from wafl.interface.dummy_interface import DummyInterface
from wafl.parsing.testcase_parser import get_user_and_bot_lines_from_text


class ConversationTestCases:
    BLUE_COLOR_START = "\033[94m"
    RED_COLOR_START = "\033[31m"
    GREEN_COLOR_START = "\033[32m"
    COLOR_END = "\033[0m"

    def __init__(self, config, text, code_path=None, logger=None):
        self._testcase_data = get_user_and_bot_lines_from_text(text)
        self._knowledge = load_knowledge(config, logger)
        self._code_path = code_path if code_path else "/"

    async def test_single_case(self, name):
        if name not in self._testcase_data:
            raise RuntimeWarning(f"The testcase '{name}' does not exist")

        user_lines = self._testcase_data[name]["user_lines"]
        test_lines = self._testcase_data[name]["lines"]
        is_negated = self._testcase_data[name]["negated"]
        interface = DummyInterface(user_lines)
        conversation_events = ConversationEvents(
            self._knowledge, interface=interface, code_path=self._code_path
        )

        print(self.BLUE_COLOR_START + f"\nRunning test '{name}'." + self.COLOR_END)

        continue_conversations = True
        while continue_conversations:
            try:
                continue_conversations = await conversation_events.process_next()

            except (IndexError, CloseConversation):
                break

        is_consistent = True
        generated_lines = interface.get_utterances_list()
        for test_line, generated_line in zip(test_lines, generated_lines):
            test_line = self._apply_deixis(test_line)
            if not await self._lhs_is_similar_to(generated_line, test_line):
                print(f" [test_line] {test_line}")
                print(f" [predicted_line] {generated_line}")
                is_consistent = False
                break

        if (is_consistent and not is_negated) or (not is_consistent and is_negated):
            print(self.GREEN_COLOR_START + " [Success]\n\n" + self.COLOR_END)
            return True

        print(self.RED_COLOR_START + " [Fail]\n\n" + self.COLOR_END)
        print("\n This is how the dialogue went:")
        for line in interface.get_utterances_list():
            print(" " + line)

        return False

    async def run(self):
        to_return = True

        for name in self._testcase_data:
            result = await self.test_single_case(name)
            if not result:
                to_return = False

        return to_return

    async def _lhs_is_similar_to(self, lhs, rhs):
        lhs_name = lhs.split(":")[0].strip()
        rhs_name = rhs.split(":")[0].strip()
        if lhs_name != rhs_name:
            return False

        lhs = ":".join(item.strip() for item in lhs.split(":")[1:])
        rhs = ":".join(item.strip() for item in rhs.split(":")[1:])
        return fuzz.ratio(lhs, rhs) > 80

    def _apply_deixis(self, line):
        name = line.split(":")[0].strip()

        if name.lower() == "user":
            return from_user_to_bot(line)

        if name.lower() == "bot":
            return from_bot_to_user(line)



import re

from nltk import pos_tag, word_tokenize


def is_question(text):
    text = text.strip()
    if not text:
        return False

    if ":" in text:
        return False

    if text[-1] == "?":
        return True

    if ", the user says" in text:
        return False

    text = re.sub("^Are", "are", text)
    text = re.sub("^Am", "am", text)

    word_and_pos_list = pos_tag(word_tokenize(text))
    first_tag = word_and_pos_list[0][1]
    if first_tag in ["VBZ", "VBD", "VBP", "MD", "WRB", "WP", "WDT"]:
        return True

    return False


def is_yes_no_question(text):
    text = text.strip()
    if not text:
        return False

    text = re.sub("^Are", "are", text)
    text = re.sub("^Am", "am", text)

    word_and_pos_list = pos_tag(word_tokenize(text))
    first_tag = word_and_pos_list[0][1]
    if first_tag in ["VBZ", "VBD", "VBP", "MD"]:
        return True

    return False


def get_sentence_from_yn_question(text):
    text = text.strip()
    text = text.replace("?", "")
    if not text:
        return ""

    triggers_list = [
        ("NN", "DT"),
        ("NNP", "DT"),
        ("NN", "JJ"),
        ("NNP", "JJ"),
        ("NN", "RB"),
        ("NNP", "RB"),
        ("NN", "VBG"),
        ("NNP", "VBG"),
        ("NN", "VB"),
        ("NN", "VBP"),
        ("NN", "VBD"),
        ("NNP", "VB"),
        ("NNP", "VBP"),
        ("NN", "VBN"),
        ("NNP", "VBN"),
        ("NNP", "VBD"),
        ("NNS", "DT"),
        ("NNS", "JJ"),
        ("NNS", "RB"),
        ("NNS", "VBG"),
        ("NNS", "VBN"),
        ("NNS", "VBD"),
        ("NNS", "VB"),
        ("NNS", "VBP"),
        ("PRP", "DT"),
        ("PRP", "JJ"),
        ("PRP", "RB"),
        ("PRP", "VBG"),
        ("PRP", "VBN"),
        ("PRP", "VBD"),
        ("PRP", "VB"),
        ("PRP", "VBP"),
    ]

    word_and_pos_list = pos_tag(word_tokenize(text))
    first_word = word_and_pos_list[0][0].lower()
    new_word_list = []
    words_has_been_added = False
    for index in range(1, len(word_and_pos_list[1:])):
        new_word_list.append(word_and_pos_list[index][0])
        curr_pos = word_and_pos_list[index][1]
        next_pos = word_and_pos_list[index + 1][1]
        if not words_has_been_added and (curr_pos, next_pos) in triggers_list:
            new_word_list.append(first_word)
            words_has_been_added = True

    if not words_has_been_added:
        new_word_list.append(first_word)

    new_word_list.append(word_and_pos_list[-1][0])
    return " ".join(new_word_list)



class ConversationalMemory:
    def __init__(self):
        self._story = ""
        self._questions = []
        self._answers = []
        self._failed_clauses = set()

    def get_story(self):
        return self._story

    def get_discussion(self):
        discussion = ""
        for question, answer in zip(self._questions, self._answers + [""]):
            discussion += "Q: " + question + "\n"
            discussion += "A: " + answer + "\n"

        return discussion

    def add_story(self, text):
        if text[-1] != ".":
            text += "."

        self._story += " " + text

    def add_question(self, text):
        self._questions.append(text)

    def add_answer(self, text):
        self._answers.append(text)

    def text_is_in_prior_questions(self, text: str):
        for question in self._questions:
            if text.lower() in question.lower():
                return True

        return False

    def text_is_in_prior_answers(self, text: str):
        for answer in self._answers:
            if text.lower() in answer.lower():
                return True

        return False

    def add_failed_clause(self, clause):
        self._failed_clauses.add(clause)

    def is_in_prior_failed_clauses(self, clause):
        return clause in self._failed_clauses

    def erase(self):
        self._story = ""
        self._questions = []
        self._answers = []
        self._failed_clauses = set()



import asyncio
import os
import random
import sys
import threading

from flask import Flask, render_template, redirect, url_for
from flask_cors import CORS
from wafl.config import Configuration
from wafl.events.conversation_events import ConversationEvents
from wafl.interface.queue_interface import QueueInterface
from wafl.knowledge.single_file_knowledge import SingleFileKnowledge
from wafl.logger.local_file_logger import LocalFileLogger
from wafl.scheduler.conversation_loop import ConversationLoop
from wafl.scheduler.scheduler import Scheduler
from wafl.scheduler.web_loop import WebLoop

_path = os.path.dirname(__file__)
_logger = LocalFileLogger()
app = Flask(
    __name__,
    static_url_path="",
    static_folder=os.path.join(_path, "../frontend/"),
    template_folder=os.path.join(_path, "../frontend/"),
)
CORS(app)


@app.route("/create_new_instance", methods=["POST"])
def create_new_instance():
    conversation_id = random.randint(0, sys.maxsize)
    result = create_scheduler_and_webserver_loop(conversation_id)
    add_new_rules(app, conversation_id, result["web_server_loop"])
    thread = threading.Thread(target=result["scheduler"].run)
    thread.start()
    return redirect(url_for(f"index_{conversation_id}"))


@app.route("/")
async def index():
    return render_template("selector.html")


def get_app():
    return app


def create_scheduler_and_webserver_loop(conversation_id):
    config = Configuration.load_local_config()
    interface = QueueInterface()
    interface.activate()
    conversation_events = ConversationEvents(
        config=config,
        interface=interface,
        logger=_logger,
    )
    conversation_loop = ConversationLoop(
        interface,
        conversation_events,
        _logger,
        activation_word="",
        max_misses=-1,
        deactivate_on_closed_conversation=False,
    )
    asyncio.run(interface.output("Hello. How may I help you?"))
    web_loop = WebLoop(interface, conversation_id, conversation_events)
    return {
        "scheduler": Scheduler([conversation_loop, web_loop]),
        "web_server_loop": web_loop,
    }


def add_new_rules(app, conversation_id, web_server_loop):
    app.add_url_rule(
        f"/{conversation_id}/",
        f"index_{conversation_id}",
        web_server_loop.index,
        methods=["GET"],
    )
    app.add_url_rule(
        f"/{conversation_id}/reset_conversation",
        f"reset_conversation_{conversation_id}",
        web_server_loop.reset_conversation,
        methods=["POST"],
    )
    app.add_url_rule(
        f"/{conversation_id}/reload_rules",
        f"reload_rules_{conversation_id}",
        web_server_loop.reload_rules,
        methods=["POST"],
    )
    app.add_url_rule(
        f"/{conversation_id}/check_new_messages",
        f"check_new_messages_{conversation_id}",
        web_server_loop.check_for_new_messages,
        methods=["POST"],
    )
    app.add_url_rule(
        f"/{conversation_id}/load_messages",
        f"load_messages_{conversation_id}",
        web_server_loop.load_messages,
        methods=["POST", "GET"],
    )
    app.add_url_rule(
        f"/{conversation_id}/input",
        f"input_{conversation_id}",
        web_server_loop.handle_input,
        methods=["POST", "GET"],
    )
    app.add_url_rule(
        f"/{conversation_id}/output",
        f"output_{conversation_id}",
        web_server_loop.handle_output,
        methods=["POST"],
    )
    app.add_url_rule(
        f"/{conversation_id}/thumbs_up",
        f"thumbs_up_{conversation_id}",
        web_server_loop.thumbs_up,
        methods=["POST"],
    )
    app.add_url_rule(
        f"/{conversation_id}/thumbs_down",
        f"thumbs_down_{conversation_id}",
        web_server_loop.thumbs_down,
        methods=["POST"],
    )



from wafl.simple_text_processing.questions import is_question
from wafl.facts import Fact
from wafl.rules import Rule


def parse_rule_from_single_line(text, knowledge_name=None):
    if ":-" not in text:
        return None

    text = text.strip()
    effect_text, causes_text = text.split(":-")
    effect = Fact(
        text=effect_text.strip(),
        is_question=is_question(effect_text),
        knowledge_name=knowledge_name,
    )
    causes = [
        Fact(
            text=item.strip(),
            is_question=is_question(item),
            knowledge_name=knowledge_name,
        )
        for item in causes_text.split(";")
    ]

    return Rule(effect=effect, causes=causes)



import hashlib
import os
import time


class LogLevels:
    INFO = 0
    WARNING = 1
    ERROR = 2


class HistoryLogger:
    level = LogLevels

    def __init__(self, interface: "BaseInterface", directory: str = "logs/"):
        if not os.path.exists(directory):
            os.mkdir(directory)

        filename = (
            hashlib.md5(str(time.time()).encode("utf8")).hexdigest() + "-history.log"
        )
        self._filename = os.path.join(directory, filename)
        self._depth = 0
        self._log_level = LogLevels.INFO
        self._interface = interface
        self._max_num_past_utterances = 7
        self._cache = set()

    def set_depth(self, depth: int):
        self._depth = depth

    def set_log_level(self, log_level: LogLevels):
        self._log_level = log_level

    def write(self, text: str, log_level=LogLevels.INFO):
        dialogue = self._interface.get_utterances_list_with_timestamp()[
            -self._max_num_past_utterances :
        ]
        start_time = -1
        if dialogue:
            start_time = dialogue[0][0]

        choices = self._interface.get_choices_and_timestamp()
        facts = self._interface.get_facts_and_timestamp()
        dialogue_items = dialogue + choices + facts
        dialogue_items = sorted(dialogue_items, key=lambda x: x[0])
        dialogue_items = [item[1] for item in dialogue_items if item[0] >= start_time]
        dialogue_items = "\n".join(dialogue_items)
        to_write = f"\n\n<{text}>\n{dialogue_items}\n</{text}>"
        if to_write in self._cache:
            return

        self._cache.add(to_write)
        with open(self._filename, "a") as f:
            f.write(to_write)



import json
import os
import shutil

_path = os.path.dirname(__file__)


def create_initial_files():
    _sample_project_dir = os.path.join(_path, "templates/")
    print("+ Initializing ... ", end="")
    shutil.copytree(_sample_project_dir, "./", dirs_exist_ok=True)

    if not os.path.exists("logs/"):
        os.mkdir("logs/")

    print("Done.")


class Configuration:
    def __init__(self, filename):
        with open(filename) as file:
            self._data = json.load(file)

    def get_value(self, key):
        if key in self._data:
            return self._data[key]

        print(
            f"The value '{key}' does not appear in the configuration. Should you add it?"
        )

    def set_value(self, key, value):
        if key not in self._data:
            raise ValueError(f"Key '{key}' does not exist in the config file")

        self._data[key] = value

    @classmethod
    def load_local_config(cls):
        try:
            return cls("config.json")

        except FileNotFoundError:
            print(
                "Cannot load 'config.json'. Does the file exist in the execution path?"
            )
            exit(0)

    @classmethod
    def load_from_filename(cls, filename):
        try:
            return cls(filename)

        except FileNotFoundError:
            print(
                f"Cannot load '{filename}'. Does the file exist in the execution path?"
            )
            exit(0)



import asyncio


class Scheduler:
    def __init__(self, loops_to_run_list):
        self._loops_to_run = [task.run() for task in loops_to_run_list]

    def run(self):
        asyncio.run(self._async_run())

    async def _async_run(self):
        await asyncio.gather(*self._loops_to_run, return_exceptions=True)



import asyncio

from wafl.config import Configuration
from wafl.exceptions import CloseConversation
from wafl.events.conversation_events import ConversationEvents
from wafl.interface.command_line_interface import CommandLineInterface
from wafl.logger.local_file_logger import LocalFileLogger
from wafl.testcases import ConversationTestCases
from wafl.variables import get_variables

_logger = LocalFileLogger()


def print_incipit():
    print()
    print(f"Running WAFL version {get_variables()['version']}.")
    print()


def run_from_command_line():
    interface = CommandLineInterface()
    config = Configuration.load_local_config()
    conversation_events = ConversationEvents(
        config=config,
        interface=interface,
        logger=_logger,
    )
    asyncio.run(interface.output("Hello. How may I help you?"))

    while True:
        try:
            asyncio.run(conversation_events.process_next())
        except (CloseConversation, KeyboardInterrupt, EOFError):
            break

    asyncio.run(interface.output("Goodbye!"))


def run_testcases():
    print("Running the testcases in testcases.txt\n")
    config = Configuration.load_local_config()
    test_cases_text = open("testcases.txt").read()
    testcases = ConversationTestCases(
        config,
        test_cases_text,
        logger=_logger,
    )
    asyncio.run(testcases.run())


def download_models():
    import nltk

    nltk.download("averaged_perceptron_tagger")



import os
import numpy as np

from typing import List, Tuple
from gensim.models import KeyedVectors
from wafl.connectors.factories.sentence_embedder_connector_factory import (
    SentenceEmbedderConnectorFactory,
)
from wafl.retriever.base_retriever import BaseRetriever

_path = os.path.dirname(__file__)


class DenseRetriever(BaseRetriever):
    _threshold_length = 5

    def __init__(self, model_name, config):
        self._connector = SentenceEmbedderConnectorFactory.get_connector(
            model_name, config
        )
        self._embeddings_model = KeyedVectors(384)

    async def add_text_and_index(self, text: str, index: str):
        embeddings = await self._get_embeddings_from_text(text)
        self._embeddings_model.add_vectors([index], [embeddings])
        self._embeddings_model.fill_norms(force=True)

    async def get_indices_and_scores_from_text(
        self, text: str
    ) -> List[Tuple[str, float]]:
        if not text or len(text) < self._threshold_length:
            return []

        embeddings = await self._get_embeddings_from_text(text)
        return self._embeddings_model.similar_by_vector(embeddings, topn=5)

    async def _get_embeddings_from_text(self, text: str) -> "numpy.array":
        return (await self._connector.predict(text))["embedding"]

    async def get_dot_product(self, lhs: str, rhs: str) -> float:
        lhs_vector = (await self._connector.predict(lhs))["embedding"]
        rhs_vector = (await self._connector.predict(rhs))["embedding"]
        return (
            np.dot(lhs_vector, rhs_vector)
            / np.linalg.norm(lhs_vector)
            / np.linalg.norm(rhs_vector)
        )



from dataclasses import dataclass
from typing import List


@dataclass
class Rule:
    effect: "Fact"
    causes: List["Fact"]
    knowledge_name: str = "/"

    def toJSON(self):
        return str(self)

    def __str__(self):
        rule_str = self.effect.text
        for cause in self.causes:
            try:
                rule_str += "\n  " + cause.text

            except TypeError as e:
                print(f"Error in rule:'''\n{rule_str}'''")
                print("Perhaps the YAML file is not well formatted.")
                print()
                raise e

        return rule_str



import re


def get_answer_from_text(text):
    _claim_yn = "The claim makes sense:\n"
    pos = text.find(_claim_yn) + len(_claim_yn)
    return text[pos]


def get_text_up_to_question(text):
    _claim_yn = "The claim makes sense:\n"
    return text[: text.find(_claim_yn) + len(_claim_yn)]


def get_function_description(text):
    if "<" not in text:
        return ""

    return re.sub(r".*<(.*)>$", r"\1", text, re.MULTILINE).strip()


def get_code(text):
    return re.sub(r"(.*)<.*>$", r"\1", text, re.MULTILINE).strip()



from dataclasses import dataclass

from wafl.simple_text_processing.questions import is_question


@dataclass
class Query:
    text: str
    is_question: bool
    variable: str = None

    def is_neutral(self) -> bool:
        return normalized(self.text) in ["unknown"]

    @staticmethod
    def create_from_text(text):
        return Query(text=text, is_question=is_question(text))


@dataclass
class Answer:
    text: str
    variable: str = None

    def is_true(self) -> bool:
        return normalized(self.text) in ["yes", "true"]

    def is_false(self) -> bool:
        return normalized(self.text) in ["no", "false"]

    def is_neutral(self) -> bool:
        return normalized(self.text) in ["unknown"]

    @staticmethod
    def create_true():
        return Answer(text="true")

    @staticmethod
    def create_false():
        return Answer(text="false")

    @staticmethod
    def create_neutral(variable=None):
        return Answer(text="unknown", variable=variable)

    @staticmethod
    def create_from_text(text):
        return Answer(text=text)


def normalized(text):
    text = text.strip()
    if not text:
        return ""

    if text[-1] == ".":
        text = text[:-1]

    return text.lower()



import time

from wafl.simple_text_processing.deixis import from_bot_to_user
from wafl.interface.base_interface import BaseInterface
from wafl.interface.utils import not_good_enough

COLOR_START = "\033[94m"
COLOR_END = "\033[0m"


class CommandLineInterface(BaseInterface):
    def __init__(self):
        super().__init__()
        self._bot_has_spoken = False

    async def output(self, text: str, silent: bool = False):
        if silent:
            print(text)
            return

        utterance = from_bot_to_user(text)
        print(COLOR_START + "bot> " + utterance + COLOR_END)
        self._utterances.append((time.time(), f"bot: {text}"))
        self.bot_has_spoken(True)

    async def input(self) -> str:
        text = input("user> ").strip()
        while not_good_enough(text):
            await self.output("I did not quite understand that")
            text = input("user> ")

        self._utterances.append((time.time(), f"user: {text}"))
        return text

    def bot_has_spoken(self, to_set: bool = None):
        if to_set != None:
            self._bot_has_spoken = to_set

        return self._bot_has_spoken



from typing import List, Tuple

from wafl.retriever.base_retriever import BaseRetriever


class StringRetriever(BaseRetriever):
    def __init__(self):
        self._string_dict = {}

    async def add_text_and_index(self, text: str, index: str):
        self._string_dict[text.strip()] = index

    async def get_indices_and_scores_from_text(
        self, text: str
    ) -> List[Tuple[str, float]]:
        if text.strip() in self._string_dict:
            return [(self._string_dict[text.strip()], 1)]

        return []



class LogLevels:
    INFO = 0
    WARNING = 1
    ERROR = 2


class BaseLogger:
    level = LogLevels

    def set_depth(self, depth: int):
        raise NotImplementedError()

    def set_log_level(self, log_level: LogLevels):
        raise NotImplementedError()

    def write(self, text: str, log_level=LogLevels.INFO):
        raise NotImplementedError



import re

from nltk import pos_tag
from nltk import word_tokenize


def from_user_to_bot(text):
    text = re.sub("are you", "is this bot", text, flags=re.IGNORECASE)
    text = re.sub("you are", "this bot is", text, flags=re.IGNORECASE)

    word_and_pos_list = pos_tag(word_tokenize(text))
    new_text = []

    for word, pos in word_and_pos_list:
        if word.lower() in ["your", "yours"] and pos == "PRP$":
            new_text.append("this bot's")

        elif word.lower() in ["you", "yourself"] and pos == "PRP":
            new_text.append("this bot")

        else:
            new_text.append(word)

    return_text = " ".join(new_text)
    return_text = return_text.replace(" @ ", "@")
    return_text = return_text.replace("does n't", "does not")
    return_text = return_text.replace("[ ", "[")
    return_text = return_text.replace(" ]", "]")
    return_text = return_text.replace("{ ", "{")
    return_text = return_text.replace(" }", "}")
    return_text = return_text.replace("``", '"')
    return_text = return_text.replace("''", '"')
    return return_text


def from_bot_to_user(text):
    text = re.sub("the user doesn't", "you don't", text, flags=re.IGNORECASE)
    text = re.sub("the user does not", "you do not", text, flags=re.IGNORECASE)
    text = re.sub("the user does", "you do", text, flags=re.IGNORECASE)
    text = re.sub("the user's", "your", text, flags=re.IGNORECASE)
    text = re.sub("does the user", "do you", text, flags=re.IGNORECASE)
    text = re.sub("is the user", "are you", text, flags=re.IGNORECASE)
    text = re.sub("the user is", "you are", text, flags=re.IGNORECASE)
    text = re.sub("the user has", "you have", text, flags=re.IGNORECASE)
    text = re.sub("the user", "you", text, flags=re.IGNORECASE)
    text = re.sub("this bot is", "I am", text, flags=re.IGNORECASE)
    text = re.sub("the bot is", "I am", text, flags=re.IGNORECASE)
    return text


def from_bot_to_bot(text):
    text = re.sub("I don't", "The bot doesn't", text, flags=re.IGNORECASE)
    text = re.sub("I do not", "The bot does not", text, flags=re.IGNORECASE)
    text = re.sub("I do not", "The bot does not", text, flags=re.IGNORECASE)
    text = re.sub(" me ", " this bot ", text, flags=re.IGNORECASE)
    text = re.sub(" my ", " this bot's ", text, flags=re.IGNORECASE)
    text = re.sub(" mine ", " this bot's ", text, flags=re.IGNORECASE)
    text = re.sub("you do", "the user does ", text, flags=re.IGNORECASE)
    text = re.sub("your", "the user's", text, flags=re.IGNORECASE)
    text = re.sub("do you", "does the user", text, flags=re.IGNORECASE)
    text = re.sub("are you", "is the user", text, flags=re.IGNORECASE)
    text = re.sub("you are", "the user is", text, flags=re.IGNORECASE)
    text = re.sub("you", "the user", text, flags=re.IGNORECASE)
    return text



import hashlib
import os.path
import time

from wafl.logger.base_logger import LogLevels, BaseLogger


class LocalFileLogger(BaseLogger):
    def __init__(self, directory: str = "logs/"):
        if not os.path.exists(directory):
            os.mkdir(directory)

        filename = hashlib.md5(str(time.time()).encode("utf8")).hexdigest() + ".log"
        self._filename = os.path.join(directory, filename)
        self._depth = 0
        self._log_level = LogLevels.INFO

    def set_depth(self, depth: int):
        self._depth = depth

    def set_log_level(self, log_level: LogLevels):
        self._log_level = log_level

    def write(self, text: str, log_level=LogLevels.INFO):
        with open(self._filename, "a") as file:
            if log_level >= self._log_level:
                file.write(" " * self._depth + text + "\n")
                file.flush()



from wafl.answerer.dialogue_answerer import DialogueAnswerer


def create_answerer(config, knowledge, interface, logger):
    return DialogueAnswerer(
        config, knowledge, interface, config.get_value("functions"), logger
    )



import json
import requests

from bs4 import BeautifulSoup
from datetime import datetime, timedelta
from wafl.exceptions import CloseConversation

_db_filename = "db.json"


def close_conversation():
    raise CloseConversation()


def check_today_weather():
    today = datetime.now().strftime("%Y-%m-%d")
    with open(_db_filename) as file:
        db = json.load(file)
        latitude = db["latitude"]
        longitude = db["longitude"]

    return check_weather_lat_long(latitude, longitude, today)


def check_tomorrow_weather():
    today = datetime.now()
    with open(_db_filename) as file:
        db = json.load(file)
        latitude = db["latitude"]
        longitude = db["longitude"]

    tomorrow = (today + timedelta(days=1)).strftime("%Y-%m-%d")
    return check_weather_lat_long(latitude, longitude, tomorrow)


def check_weather_lat_long(latitude, longitude, day):
    secrets = json.load(open("secrets.json"))
    result = requests.get(
        f"https://rgw.5878-e94b1c46.eu-gb.apiconnect.appdomain.cloud/metoffice/production/v0/forecasts/point/daily?excludeParameterMetadata=true&includeLocationName=true&latitude={latitude}&longitude={longitude}",
        headers={
            "X-IBM-Client-Id": secrets["key"],
            "X-IBM-Client-Secret": secrets["secret"],
        },
    )
    data = result.json()
    if "features" not in data:
        return "There is a connection error to the weather API. Please try later. "
    to_say = ""
    for item in data["features"][0]["properties"]["timeSeries"]:
        if day in item["time"]:
            to_say += f"The temperature should be between {int(item['dayLowerBoundMaxTemp'])} and {int(item['dayUpperBoundMaxTemp'])}. "
            if item["dayProbabilityOfPrecipitation"] != 0:
                to_say += f"The probability of rain is {item['dayProbabilityOfPrecipitation']} percent. "

            else:
                to_say += "There is no probability of rain. "

            if item["dayProbabilityOfSnow"] != 0:
                to_say += f"The probability of snow is {item['dayProbabilityOfSnow']} percent. "

            else:
                to_say += "There is no probability of snow. "

    return to_say


def get_website(url):
    response = requests.get(url)
    soup = BeautifulSoup(response.content, "html.parser")
    text = soup.get_text()
    return text


def get_time():
    return datetime.now().strftime("%H:%M")


def get_date():
    return datetime.now().strftime("%Y-%m-%d")


def get_day():
    return datetime.now().strftime("%A")


def add_to_shopping_list(list_of_items_to_add):
    db = json.load(open(_db_filename))
    for item in list_of_items_to_add:
        if item not in db["shopping_list"]:
            db["shopping_list"].append(item)

    json.dump(db, open(_db_filename, "w"))

    return "Item added"


def remove_shopping_list(list_of_items_to_remove):
    db = json.load(open(_db_filename))
    for item in list_of_items_to_remove:
        if item in db["shopping_list"]:
            db["shopping_list"].remove(item)

    json.dump(db, open(_db_filename, "w"))

    return "Item removed"


def get_shopping_list():
    db = json.load(open(_db_filename))
    if db["shopping_list"] == []:
        return "nothing"

    return ", ".join(db["shopping_list"])



import yaml

from wafl.facts import Fact
from wafl.rules import Rule
from wafl.simple_text_processing.deixis import from_user_to_bot


def get_facts_and_rules_from_text(text: str, knowledge_name=None):
    parsed_text_dict = yaml.safe_load(text)
    fact_strings = parsed_text_dict.get("facts", [])
    rules_list = parsed_text_dict.get("rules", {})

    facts = []
    for text in fact_strings:
        facts.append(
            Fact(
                text=from_user_to_bot(text),
            )
        )

    rules = []
    for rule_dict in rules_list:
        rules.append(
            Rule(
                effect=Fact(text=list(rule_dict.keys())[0]),
                causes=[Fact(item) for item in list(rule_dict.values())[0]],
            )
        )

    return {"facts": facts, "rules": rules}



from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import ENGLISH_STOP_WORDS
from wafl.simple_text_processing.normalize import normalized


def get_most_common_words(text, max_num_words, count_threshold=1):
    corpus = text.split("\n")
    cv = CountVectorizer(ngram_range=(1, 2))
    words_count = cv.fit_transform(corpus).sum(axis=0)
    words_freq = [
        (word, words_count[0, index])
        for word, index in cv.vocabulary_.items()
        if words_count[0, index] >= count_threshold
        and word not in ENGLISH_STOP_WORDS
        and "user" not in word
        and "bot" not in word
    ]
    words_freq = sorted(words_freq, key=lambda x: -x[1])
    return [word for word, _ in words_freq][:max_num_words]


def not_good_enough(text):
    if not text:
        return True

    if "[unclear]" in text:
        return True

    if normalized(text) != "no" and len(text.strip().replace(" ", "")) < 3:
        return True

    return False



from wafl.variables import get_variables
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="wafl",
    version=get_variables()["version"],
    url="http://github.com/fractalego/wafl",
    author="Alberto Cetoli",
    author_email="alberto@fractalego.io",
    description="A hybrid chatbot.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=[
        "wafl",
        "wafl.answerer",
        "wafl.connectors",
        "wafl.connectors.bridges",
        "wafl.connectors.factories",
        "wafl.connectors.remote",
        "wafl.events",
        "wafl.extractors",
        "wafl.filter",
        "wafl.inference",
        "wafl.interface",
        "wafl.knowledge",
        "wafl.listener",
        "wafl.logger",
        "wafl.parsing",
        "wafl.retriever",
        "wafl.runners",
        "wafl.scheduler",
        "wafl.simple_text_processing",
        "wafl.speaker",
    ],
    package_data={
        "wafl": [
            "templates/*",
            "sounds/*",
            "models/*",
            "frontend/*",
        ],
    },
    install_requires=[
        "flask[async]==2.0.1",
        "flask-cors==3.0.10",
        "flask_dropzone==1.6.0",
        "werkzeug==2.1.2",
        "nltk==3.6.2",
        "gensim==4.3.1",
        "sklearn==0.0",
        "python-Levenshtein==0.12.2",
        "wave==0.0.2",
        "fuzzywuzzy==0.18.0",
        "PyAudio==0.2.13",
        "pyctcdecode==0.2.1",
        "num2words==0.5.12",
        "word2number==1.1",
        "aiohttp==3.8.4",
        "einops==0.6.1",
        "g2p-en==2.1.0",
        "pyyaml==6.0.1",
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
    ],
    entry_points={
        "console_scripts": ["wafl=wafl.command_line:main"],
    },
    include_package_data=True,
    zip_safe=False,
)



import wafl.simple_text_processing.questions
from wafl.simple_text_processing.questions import get_sentence_from_yn_question


def text_is_exact_string(text):
    return text.strip() and text.strip()[0] == "_"


async def rules_are_too_different(retriever, rules):
    dot_products = []
    for item in rules[1:]:
        dot_products.append(
            await retriever.get_dot_product(item.effect.text, rules[0].effect.text)
        )

    if dot_products and min(dot_products) < 0.39:
        return False

    return True


def get_first_cluster_of_rules(rules_and_threshold):
    if not rules_and_threshold:
        return []

    _cluster_margin = 0.1

    last_threshold = rules_and_threshold[0][1]
    prior_rules = set()
    rules = []
    for rule, threshold in rules_and_threshold:
        if str(rule) in prior_rules:
            continue

        prior_rules.add(str(rule))
        if abs(threshold - last_threshold) < _cluster_margin:
            rules.append(rule)

        else:
            break

    return rules


def filter_out_rules_that_are_too_dissimilar_to_query(query, rules_and_scores):
    num_query_words = len(query.text.split())
    new_rules_and_scores = []
    for item in rules_and_scores:
        rule = item[0]
        num_rule_effect_words = len(rule.effect.text.split())
        if num_query_words < num_rule_effect_words / 3:
            continue

        new_rules_and_scores.append(item)

    return new_rules_and_scores


async def filter_out_rules_through_entailment(entailer, query, rules_and_scores):
    new_rules_and_scores = []
    for rule, score in rules_and_scores:
        if rule.effect.is_question:
            new_rules_and_scores.append((rule, score))

        elif needs_substitutions(rule.effect):
            new_rules_and_scores.append((rule, score))

        else:
            entailment_score = await entailer.entails(
                query.text, rule.effect.text, return_threshold=True
            )
            if entailment_score:
                new_rules_and_scores.append((rule, score * entailment_score))
                continue

            entailment_score = await entailer.entails(
                rule.effect.text, query.text, return_threshold=True
            )
            if entailment_score:
                new_rules_and_scores.append((rule, score * entailment_score))

    return sorted(new_rules_and_scores, key=lambda x: -x[1])


def needs_substitutions(effect):
    if any(item in effect.text for item in ["{", "}"]):
        return True

    return False



import asyncio
import time

from wafl.interface.base_interface import BaseInterface


class QueueInterface(BaseInterface):
    def __init__(self, output_filter=None):
        super().__init__()
        self._bot_has_spoken = False
        self.input_queue = []
        self.output_queue = []
        self._output_filter = output_filter

    async def output(self, text: str, silent: bool = False):
        if silent:
            self.output_queue.append({"text": text, "silent": True})
            return

        if self._output_filter:
            text = await self._output_filter.filter(
                self.get_utterances_list_with_timestamp(), text
            )

        utterance = text
        self.output_queue.append({"text": utterance, "silent": False})
        self._utterances.append((time.time(), f"bot: {text}"))
        self.bot_has_spoken(True)

    async def input(self) -> str:
        while not self.input_queue:
            await asyncio.sleep(0.1)

        text = self.input_queue.pop(0)
        self._utterances.append((time.time(), f"user: {text}"))
        return text

    def bot_has_spoken(self, to_set: bool = None):
        if to_set != None:
            self._bot_has_spoken = to_set

        return self._bot_has_spoken



import asyncio
import logging
from typing import List

import nltk

from wafl.config import Configuration
from wafl.facts import Fact
from wafl.simple_text_processing.normalize import normalized
from wafl.knowledge.base_knowledge import BaseKnowledge
from wafl.knowledge.utils import (
    text_is_exact_string,
    get_first_cluster_of_rules,
    filter_out_rules_that_are_too_dissimilar_to_query,
)
from wafl.parsing.line_rules_parser import parse_rule_from_single_line
from wafl.parsing.rules_parser import get_facts_and_rules_from_text
from wafl.extractors.dataclasses import Query
from wafl.retriever.string_retriever import StringRetriever
from wafl.retriever.dense_retriever import DenseRetriever
from wafl.text_utils import clean_text_for_retrieval

nltk.download("punkt")
nltk.download("averaged_perceptron_tagger")

_logger = logging.getLogger(__name__)


class SingleFileKnowledge(BaseKnowledge):
    _threshold_for_questions_from_user = 0.55
    _threshold_for_questions_from_bot = 0.6
    _threshold_for_questions_in_rules = 0.49
    _threshold_for_facts = 0.4
    _threshold_for_fact_rules = 0.22
    _threshold_for_fact_rules_for_creation = 0.1
    _threshold_for_partial_facts = 0.48
    _max_rules_per_type = 3

    def __init__(self, config, rules_text=None, knowledge_name=None, logger=None):
        self._logger = logger
        self._facts_dict = {}
        self._rules_dict = {}
        self._facts_retriever = DenseRetriever("text_embedding_model", config)
        self._facts_retriever_for_questions = DenseRetriever(
            "text_embedding_model",
            config,
        )
        self._rules_incomplete_retriever = DenseRetriever(
            "text_embedding_model", config
        )
        self._rules_fact_retriever = DenseRetriever("text_embedding_model", config)
        self._rules_question_retriever = DenseRetriever("text_embedding_model", config)
        self._rules_string_retriever = StringRetriever()
        knowledge_name = knowledge_name if knowledge_name else self.root_knowledge
        if rules_text:
            facts_and_rules = get_facts_and_rules_from_text(
                rules_text, knowledge_name=knowledge_name
            )
            self._facts_dict = {
                f"F{index}": value
                for index, value in enumerate(facts_and_rules["facts"])
            }
            self._rules_dict = {
                f"R{index}": value
                for index, value in enumerate(facts_and_rules["rules"])
            }
            try:
                loop = asyncio.get_running_loop()

            except RuntimeError:
                loop = None

            if not loop or not loop.is_running():
                asyncio.run(self._initialize_retrievers())

    async def add(self, text, knowledge_name="/"):
        fact_index = f"F{len(self._facts_dict)}"
        self._facts_dict[fact_index] = Fact(text=text, knowledge_name=knowledge_name)
        await self._facts_retriever.add_text_and_index(
            clean_text_for_retrieval(text), fact_index
        )
        await self._facts_retriever_for_questions.add_text_and_index(
            clean_text_for_retrieval(text), fact_index
        )

    async def add_rule(self, rule_text, knowledge_name=None):
        rule = parse_rule_from_single_line(rule_text, knowledge_name)
        index = str(len(self._rules_dict))
        index = f"R{index}"
        self._rules_dict[index] = rule
        await self._rules_fact_retriever.add_text_and_index(
            clean_text_for_retrieval(rule.effect.text), index=index
        )

    async def has_better_match(self, query_text: str) -> bool:
        if any(normalized(query_text).find(item) == 0 for item in ["yes", "no"]):
            return False

        if any(normalized(query_text).find(item) != -1 for item in [" yes ", " no "]):
            return False

        rules = await self.ask_for_rule_backward(
            Query(text=f"The user says to the bot: '{query_text}.'", is_question=False)
        )
        return any(rule.effect.is_interruption for rule in rules)

    async def ask_for_facts(
        self, query, is_from_user=False, knowledge_name=None, threshold=None
    ):
        if query.is_question:
            indices_and_scores = await self._facts_retriever_for_questions.get_indices_and_scores_from_text(
                query.text
            )

        else:
            indices_and_scores = (
                await self._facts_retriever.get_indices_and_scores_from_text(query.text)
            )
        if threshold == None and is_from_user:
            threshold = (
                self._threshold_for_questions_from_user
                if query.is_question
                else self._threshold_for_facts
            )
        elif threshold == None:
            threshold = (
                self._threshold_for_questions_from_bot
                if query.is_question
                else self._threshold_for_facts
            )

        return [
            self._facts_dict[item[0]]
            for item in indices_and_scores
            if item[1] > threshold
        ]

    async def ask_for_facts_with_threshold(
        self, query, is_from_user=False, knowledge_name=None, threshold=None
    ):
        if query.is_question:
            indices_and_scores = await self._facts_retriever_for_questions.get_indices_and_scores_from_text(
                query.text
            )

        else:
            indices_and_scores = (
                await self._facts_retriever.get_indices_and_scores_from_text(query.text)
            )

        if threshold == None and is_from_user:
            threshold = (
                self._threshold_for_questions_from_user
                if query.is_question
                else self._threshold_for_facts
            )

        elif threshold == None:
            threshold = (
                self._threshold_for_questions_from_bot
                if query.is_question
                else self._threshold_for_facts
            )

        return [
            (self._facts_dict[item[0]], item[1])
            for item in indices_and_scores
            if item[1] > threshold
        ]

    async def ask_for_rule_backward(self, query, knowledge_name=None, first_n=None):
        rules_and_scores = await self._ask_for_rule_backward_with_scores(
            query, knowledge_name, first_n
        )
        return get_first_cluster_of_rules(rules_and_scores)

    def get_facts_and_rule_as_text(self):
        text = ""
        for fact in self._facts_dict.values():
            text += fact.text + "\n"

        for effect in self._rules_dict.values():
            text += effect.effect.text + "\n"

        return text

    async def _initialize_retrievers(self):
        for index, fact in self._facts_dict.items():
            if text_is_exact_string(fact.text):
                continue

            await self._facts_retriever.add_text_and_index(
                clean_text_for_retrieval(fact.text), index
            )

            await self._facts_retriever_for_questions.add_text_and_index(
                clean_text_for_retrieval(fact.text), index
            )

        for index, rule in self._rules_dict.items():
            if text_is_exact_string(rule.effect.text):
                continue

            if "{" in rule.effect.text:
                await self._rules_incomplete_retriever.add_text_and_index(
                    clean_text_for_retrieval(rule.effect.text), index
                )
                continue

            elif rule.effect.is_question:
                await self._rules_question_retriever.add_text_and_index(
                    clean_text_for_retrieval(rule.effect.text), index
                )

            else:
                await self._rules_fact_retriever.add_text_and_index(
                    clean_text_for_retrieval(rule.effect.text), index
                )

        for index, rule in self._rules_dict.items():
            if not text_is_exact_string(rule.effect.text):
                continue

            await self._rules_string_retriever.add_text_and_index(
                rule.effect.text, index
            )

    @staticmethod
    async def create_from_list(
        facts: List[str], config: "Configuration" = None
    ) -> "SingleFileKnowledge":
        if not config:
            config = Configuration.load_local_config()

        knowledge = SingleFileKnowledge(config)
        for index, fact in enumerate(facts):
            await knowledge.add(fact)

        return knowledge

    async def _ask_for_rule_backward_with_scores(
        self, query, knowledge_name=None, first_n=None
    ):
        if text_is_exact_string(query.text):
            indices_and_scores = (
                await self._rules_string_retriever.get_indices_and_scores_from_text(
                    query.text
                )
            )
            return [(self._rules_dict[item[0]], item[1]) for item in indices_and_scores]

        indices_and_scores = (
            await self._rules_fact_retriever.get_indices_and_scores_from_text(
                query.text
            )
        )
        if not first_n:
            fact_rules = [
                (self._rules_dict[item[0]], item[1])
                for item in indices_and_scores
                if item[1] > self._threshold_for_fact_rules
            ]

        else:
            fact_rules = [
                (self._rules_dict[item[0]], item[1])
                for item in indices_and_scores
                if item[1] > self._threshold_for_fact_rules_for_creation
            ]

        fact_rules = [item for item in sorted(fact_rules, key=lambda x: -x[1])][
            : self._max_rules_per_type
        ]

        indices_and_scores = (
            await self._rules_question_retriever.get_indices_and_scores_from_text(
                query.text
            )
        )
        question_rules = [
            (self._rules_dict[item[0]], item[1])
            for item in indices_and_scores
            if item[1] > self._threshold_for_questions_in_rules
        ]
        question_rules = [item for item in sorted(question_rules, key=lambda x: -x[1])][
            : self._max_rules_per_type
        ]

        indices_and_scores = (
            await self._rules_incomplete_retriever.get_indices_and_scores_from_text(
                query.text
            )
        )
        incomplete_rules = [
            (self._rules_dict[item[0]], item[1])
            for item in indices_and_scores
            if item[1] > self._threshold_for_partial_facts
        ]
        incomplete_rules = [
            item for item in sorted(incomplete_rules, key=lambda x: -x[1])
        ][: self._max_rules_per_type]

        rules_and_scores = fact_rules + question_rules + incomplete_rules
        rules_and_scores = filter_out_rules_that_are_too_dissimilar_to_query(
            query, rules_and_scores
        )
        return rules_and_scores



from wafl.connectors.remote.remote_whisper_connector import RemoteWhisperConnector


class WhisperConnectorFactory:
    @staticmethod
    def get_connector(config):
        return RemoteWhisperConnector(
            config.get_value("listener_model")
        )



import pyaudio
import wave

from wafl.speaker.base_speaker import BaseSpeaker


class SoundFileSpeaker(BaseSpeaker):
    _chunk = 1024

    def __init__(self):
        self._p = pyaudio.PyAudio()

    def speak(self, filename):
        self.__play_sound(filename)

    def __play_sound(self, sound_filename):
        f = wave.open(sound_filename, "rb")
        stream = self._p.open(
            format=self._p.get_format_from_width(f.getsampwidth()),
            channels=f.getnchannels(),
            rate=f.getframerate(),
            output=True,
        )
        data = f.readframes(self._chunk)

        while data:
            stream.write(data)
            data = f.readframes(self._chunk)

        stream.stop_stream()
        stream.close()



from unittest import TestCase

from wafl.simple_text_processing.deixis import from_user_to_bot, from_bot_to_user


class TestDeixis(TestCase):
    def test_from_user_to_bot(self):
        translation = from_user_to_bot("You are called computer")
        expected = "this bot is called computer"
        assert translation == expected

    def test_from_bot_to_user(self):
        translation = from_bot_to_user("The user is Bob")
        expected = "you are Bob"
        assert translation == expected



import os

from wafl.connectors.factories.llm_connector_factory import LLMConnectorFactory

_path = os.path.dirname(__file__)


class LLMChitChatAnswerBridge:
    def __init__(self, config):
        self._connector = LLMConnectorFactory.get_connector(config)
        self._config = config

    async def get_answer(self, text: str, dialogue: str, query: str) -> str:
        prompt = await self._get_answer_prompt(text, query, dialogue)
        return await self._connector.generate(prompt)

    async def _get_answer_prompt(self, text, rules_text, dialogue=None):
        if rules_text:
            rules_to_use = f"I want you to follow these rules:\n{rules_text.strip()}\n"
            pattern = "\nuser: "
            if pattern in dialogue:
                last_user_position = dialogue.rfind(pattern)
                before_user_dialogue, after_user_dialogue = (
                    dialogue[:last_user_position],
                    dialogue[last_user_position + len(pattern) :],
                )
                dialogue = f"{before_user_dialogue}\nuser: {rules_to_use}\nuser: {after_user_dialogue}"
            else:
                dialogue = f"user: {rules_to_use}\n{dialogue}"

        prompt = f"""
The following is a summary of a conversation. All the elements of the conversation are described briefly:
<summary>        
A user is chatting with a bot. The chat is happening through a web interface. The user is typing the messages and the bot is replying.
{text.strip()}
</summary>

<instructions>
Create a plausible dialogue based on the aforementioned summary and rules. 
Do not repeat yourself. Be friendly but not too servile.
Wrap any code or html you output in the with the markdown syntax for code blocks (i.e. use triple backticks ```) unless it is between <execute> tags.
</instructions>

This is the dialogue:
{dialogue}
bot:
        """.strip()
        return prompt



import asyncio
import pyaudio

from wafl.connectors.factories.speaker_connector_factory import SpeakerConnectorFactory
from wafl.speaker.base_speaker import BaseSpeaker
from wafl.speaker.utils import convert_numbers_to_words


class FairSeqSpeaker(BaseSpeaker):
    def __init__(self, config):
        self._connector = SpeakerConnectorFactory.get_connector(config)
        self._p = pyaudio.PyAudio()

    async def speak(self, text):
        text = convert_numbers_to_words(text)
        prediction = await self._connector.predict(text)
        wav = prediction["wav"]
        rate = prediction["rate"]
        stream = self._p.open(
            format=pyaudio.paFloat32,
            channels=1,
            rate=rate,
            output=True,
        )
        await asyncio.sleep(0.2)
        stream.write(wav)
        stream.stop_stream()
        stream.close()
        await asyncio.sleep(0.1)



def get_last_bot_utterances(dialogue_items, num_utterances):
    utterances = []
    for item in reversed(dialogue_items):
        if item[1].startswith("bot:"):
            utterances.append(item[1].replace("bot:", "").strip())

        if len(utterances) == num_utterances:
            break

    return utterances


def get_last_user_utterance(dialogue_items):
    for item in reversed(dialogue_items):
        if item[1].startswith("user:"):
            return item[1].replace("user:", "").strip()

    return ""



import asyncio

from unittest import TestCase

from wafl.config import Configuration
from wafl.events.conversation_events import ConversationEvents
from wafl.interface.dummy_interface import DummyInterface

wafl_example = """
rules:
  - the user says "my name is bob":
    - You must answer the user by writing "the horse is tall"
    
  - the user says their name:
    - reply casually to the conversation"
"""


class TestRules(TestCase):
    def test__rules_can_be_triggered(self):
        interface = DummyInterface(
            to_utter=[
                "my name is bob",
            ]
        )
        config = Configuration.load_local_config()
        config.set_value("rules", wafl_example)
        conversation_events = ConversationEvents(
            config=config,
            interface=interface,
        )
        asyncio.run(conversation_events.process_next())
        expected = "bot: the horse is tall"
        self.assertEqual(expected, interface.get_utterances_list()[-1])

    def test__rules_are_not_always_triggered(self):
        interface = DummyInterface(
            to_utter=[
                "my name is Frank",
            ]
        )
        config = Configuration.load_local_config()
        config.set_value("rules", wafl_example)
        conversation_events = ConversationEvents(
            config=config,
            interface=interface,
        )
        asyncio.run(conversation_events.process_next())
        unexpected = "bot: the horse is tall"
        self.assertNotEqual(unexpected, interface.get_utterances_list()[-1])



import re

from typing import List, Dict, Tuple, Any
from fuzzywuzzy import process
from wafl.extractors.dataclasses import Answer
from wafl.simple_text_processing.normalize import normalized
from wafl.simple_text_processing.questions import is_question


def cause_is_negated(cause_text: str) -> bool:
    return cause_text.find("!") == 0


def check_negation(cause_text: str) -> Tuple[str, bool]:
    invert_results = False
    if cause_is_negated(cause_text):
        if cause_text[0] == "!":
            invert_results = True
            cause_text = cause_text[1:]

    return cause_text, invert_results


def text_is_code(text: str) -> bool:
    if "(" in text:
        return True

    return False


def _make_safe(text: str) -> str:
    text = str(text)
    text = text.replace('"', "'")
    text = text.replace("''", "'")
    text = text.replace('""', '"')
    text = text.replace(r"\"", '\\"')
    text = text.replace(r"\'", "\\'")
    return text


def apply_substitutions(cause_text: str, substitutions: Dict[str, str]) -> str:
    if text_is_code(cause_text):
        cause_text = cause_text.replace(" ", "")

    for key, value in substitutions.items():
        safe_value = _make_safe(str(value))
        cause_text = cause_text.replace(key, safe_value)

    return cause_text


def text_has_say_command(text: str) -> bool:
    words = text.strip().split()
    if words:
        return words[0].lower() == "say"

    return False


def text_has_retrieve_command(text: str) -> bool:
    words = text.split("=")[-1].strip().split()
    if words:
        return words[0].lower() == "retrieve"

    return False


def text_has_remember_command(text: str) -> bool:
    return normalized(text).find("remember") == 0


def text_has_new_task_memory_command(text: str) -> bool:
    return normalized(text).find("erase memory") == 0


def update_substitutions_from_answer(answer: "Answer", substitutions: Dict[str, str]):
    safe_value = _make_safe(answer.text)
    substitutions[f"({{{answer.variable.strip()}}})"] = f'("{safe_value}")'
    substitutions[f"({{{answer.variable.strip()}}},"] = f'("{safe_value}",'
    substitutions[f",{{{answer.variable.strip()}}},"] = f',"{safe_value}",'
    substitutions[f",{{{answer.variable.strip()}}})"] = f',"{safe_value}")'
    substitutions[f"{{{answer.variable.strip()}}}"] = safe_value
    substitutions[f"({answer.variable.strip()})"] = f'("{safe_value}")'
    substitutions[f"({answer.variable.strip()},"] = f'("{safe_value}",'
    substitutions[f",{answer.variable.strip()},"] = f',"{safe_value}",'
    substitutions[f",{answer.variable.strip()})"] = f',"{safe_value}")'


def add_function_arguments(text: str) -> str:
    text = re.sub(
        "(.*\([\"'0-9a-zA-Z@?':\-_\.,\s]+)\)$", "\\1, self, policy, task_memory)", text
    )
    text = re.sub("(.*)\(\)$", "\\1(self, policy, task_memory)", text)
    return text


def update_substitutions_from_results(
    result: str, variable: str, substitutions: Dict[str, str]
):
    safe_value = _make_safe(result)
    substitutions.update({f"({{{variable}}})": f'("{safe_value}")'})
    substitutions.update({f"({{{variable}}},": f'("{safe_value}",'})
    substitutions.update({f",{{{variable}}},": f',"{safe_value}",'})
    substitutions.update({f",{{{variable}}})": f',"{safe_value}")'})
    substitutions.update({f"{{{variable}}}": safe_value})
    substitutions.update({f"({variable})": f'("{safe_value}")'})
    substitutions.update({f"({variable},": f'("{safe_value}",'})
    substitutions.update({f",{variable},": f',"{safe_value}",'})
    substitutions.update({f",{variable})": f',"{safe_value}")'})


def invert_answer(answer: "Answer") -> "Answer":
    if answer.text == "False":
        return Answer(text="True")

    if answer.text == "True":
        return Answer(text="False")

    if answer.text == "No":
        return Answer(text="Yes")

    if answer.text == "Yes":
        return Answer(text="No")

    return answer


def text_has_assigmnent(cause_text: str) -> bool:
    return "=" in cause_text


def process_unknown_answer(answer: "Answer") -> bool:
    if normalized(answer.text) == "unknown":
        answer = None

    return answer


def cluster_facts(facts_and_threshold: List[Tuple["Fact", float]]) -> List[str]:
    if not facts_and_threshold:
        return []

    _cluster_margin = 0.1

    texts = []
    last_threshold = facts_and_threshold[0][1]
    text = ""
    prior_facts = set()
    for fact, threshold in facts_and_threshold:
        if not fact.text:
            continue

        if fact.text in prior_facts:
            continue

        prior_facts.add(fact.text)

        if abs(threshold - last_threshold) < _cluster_margin:
            text += normalized(fact.text, lower_case=False).capitalize() + ". "

        else:
            texts.append(text.strip())
            text = ""
            last_threshold = threshold

    if text:
        texts.append(text.strip())

    return texts


def selected_answer(candidate_answers: List["Answer"], variable_name: str) -> bool:
    for answer in candidate_answers:
        if answer and normalized(answer.text) != "unknown":
            return answer

    for answer in candidate_answers:
        if answer:
            return answer

    return_answer = Answer.create_neutral()
    return_answer.variable = variable_name
    return return_answer


def fact_relates_to_user(text: str) -> bool:
    if "the user" in normalized(text):
        return True

    return False


def project_answer(answer: "Answer", candidates: List[str]) -> "Answer":
    if not candidates or not answer_is_informative(answer):
        return Answer(text="unknown")

    extracted, score = process.extract(answer.text, candidates, limit=1)[0]
    if score < 60:
        return Answer(text="unknown")

    return Answer(text=extracted)


def answer_is_informative(answer: "Answer") -> bool:
    return not any(item == normalized(answer.text) for item in ["unknown"])


def is_inference_task(text):
    text = text.split("=")[-1]
    prompt = "the user"
    if text.strip().lower().find(prompt) == 0:
        return True

    return False


async def text_is_text_generation_task(
    text: str,
    entailer: "Entailer",
) -> bool:
    if not "=" in text:
        return False

    if text_is_code(text):
        return False

    if is_inference_task(text):
        return False

    if is_question(text.split("=")[1]):
        return False

    if await entailer.entails(
        text,
        f"An instruction or request of some kind",
        return_threshold=True,
        threshold=0.5,
    ):
        return True

    return False


def escape_characters(text: str) -> str:
    text = text.replace("\n", "\\n")
    return text


def string_is_python_list(text: str) -> bool:
    try:
        result = eval(text)
        if type(result) == list:
            return True

    except (SyntaxError, NameError):
        try:
            text = text.replace("'s", "\\'s")
            result = eval(text)
            if type(result) == list:
                return True

        except (SyntaxError, NameError):
            pass

        pass

    return False


def get_list_from_string(text: str) -> List[Any]:
    try:
        result = eval(text)
        if type(result) == list:
            return result

    except (SyntaxError, NameError):
        text = text.replace("'s", "\\'s")
        result = eval(text)
        if type(result) == list:
            return result

        pass

    return []


def get_list_like_element(text: str) -> str:
    start_pos = text.find("[")
    if start_pos == -1:
        return ""

    end_pos = text.find("]", start_pos)
    if end_pos == -1:
        return ""

    return text[start_pos : end_pos + 1]


def create_python_list_from_bullet_list(text: str) -> List[str]:
    new_list = text.split("\n")
    if len(new_list) != 1:
        return new_list

    new_list = text.split("', '")
    if len(new_list) != 1:
        return new_list

    new_list = text.split(",")
    if len(new_list) != 1:
        return new_list

    return [text]


def get_causes_list(text: str) -> List[str]:
    """
    If the text contains a Python list (e.g. generate a chapter with the theme ["space", "romance"] )
    then the output becomes a list of texts where each item appears separaterly
    (e.g. ["generate a chapter with the theme space", generate a chapter with the theme romance"]

    :param text:
    :return: a list of causes,
    """
    if text_is_code(text):
        return [text]

    list_like_element = get_list_like_element(text)
    if not list_like_element:
        return [text]

    if not string_is_python_list(list_like_element):
        items = create_python_list_from_bullet_list(list_like_element[1:-1])

    else:
        items = get_list_from_string(list_like_element)

    causes_list = []
    for item in items:
        causes_list.append(text.replace(list_like_element, item))

    return causes_list


def create_default_substitutions(interface: "BaseInterface") -> Dict[str, str]:
    substitutions = {}
    dialogue = " ".join(interface.get_utterances_list())
    update_substitutions_from_results(dialogue, "_dialogue", substitutions)
    return substitutions



from wafl.connectors.remote.remote_sentence_embedder_connector import (
    RemoteSentenceEmbedderConnector,
)


class SentenceEmbedderConnectorFactory:
    @staticmethod
    def get_connector(model_name, config):
        return RemoteSentenceEmbedderConnector(config.get_value(model_name))
